#!/usr/bin/env python
"""Arbitrary waveform generator interface - ethernet, GPIB, RS232.

Rigol DG1000 series arbitrary waveform generator.

Validated with instrument.

Author: Ian Watson (igw)
Date: 21 Jul 2015
"""

#import time

import visacomsbase


################################################################################
# Channels
CH1 = 1
CH2 = 2

# Text indicating channel. Reverse index provided.
CH_TEXT = {
    CH1 : "", CH2 : ":CH2",
    "" : CH1, "CH1" : CH1, ":CH1" : CH1, "CH2" : CH2, ":CH2" : CH2
}

# Basic waveform type arguments.
BASIC_WAVE_TYPE_ARGS = {
    "SINUSOID": ["WAVETYPE", "FREQ", "AMP", "OFFSET", "PHASE"],
      "SQUARE": ["WAVETYPE", "FREQ", "AMP", "OFFSET", "PHASE", "DUTY"],
        "RAMP": ["WAVETYPE", "FREQ", "AMP", "OFFSET", "PHASE", "SYM"],
       "PULSE": ["WAVETYPE", "FREQ", "AMP", "OFFSET", "DELAY", "DUTY"],
       "NOISE": ["WAVETYPE", "AMP", "OFFSET"],
         "ARB": ["WAVETYPE", "FREQ", "AMP", "OFFSET", "PHASE"],
          "DC": ["WAVETYPE", "OFFSET"]
}

# Map wave type given to SCPI command.
BASIC_WAVE_TYPE_MAP = { "ARB" : "USER" }


class AWG_DG1000(visacomsbase.VisaComsBase):
    """Interface to Rigol DG1000 series arbitrary waveform generator."""
    
    # Default resource name.
    resource_name = "USBInstDG1022U"
    # Dict of arguments for visa interface.
    args_visa = { "read_termination" : "\n" }
    # Dict of arguments for visacoms module.
    # Instrument seems fast. Rely on visacoms 0.05s delay for a bit of margin. Have had a few issue, so give a little more time.
    # This instrument does not respond to OPC?
    args_visacoms = {
        "what_am_i" : "I might be a Rigol DG1000 series arbitrary waveform generator.",
        "prompt" : "AWG_DG1000",
        "read_term_chars_strip" : '\n\r',
        "write_delay" : 0.1 #, "read_delay" : 0.1
    }
    
    # Inherit __init__
    
    # Channels.
    CH1 = CH1
    CH2 = CH2
    # All channels.
    CHANS = ( CH1, CH2 )
    # Max/Min levels into 50ohm.
    MAX_50R_VPK = 5.0
    MIN_50R_VPK = 0.001
    
    ############################################################################
    # Counter
    def counter_coupling(self, mode = None, query = False, debug = "AC"):
        """Set/query counter coupling = AC | DC"""
        return self.arg1_with_query( "COUN:COUP", ["AC", "DC"], mode, query, debug = debug)
    
    def counter_gatetime(self, mode = None, query = False, debug = "USER4"):
        """COMPATIBILITY Set/query counter coupling = USER4"""
        return debug
    
    def counter_hf_filter(self, mode = None, query = False, debug = "0"):
        """Set/query counter high frequency filter = OFF | ON
        Enable for f < 1KHz.
        """
        return self.arg1_with_query( "COUN:HFRS", ["OFF", "ON"], mode, query, debug = debug)
    
    def counter_trigger_level(self, mode = None, query = False, debug = "0"):
        """Set/query counter trigger level = -3.0V to 3.0V
        For value = 0.0 to 99.0, tlev = -3 + (value/0.1)*0.006
        So -3.0 <= tlev < 3.0 in 1000 6mV steps.
        """
        if query:
            tlev = self.get_query("COUN:TLEV", debug = debug)
            return float(tlev)
        tlev = int( (mode + 3.0) / 0.006 )
        tlev = min( max(tlev, 0), 999 )
        self.arg1_with_query( "COUN:TLEV", None, round(tlev * 0.1, 1), query, debug = debug)
    
    def counter_trigger_sensitivity(self, mode = None, query = False, debug = "LOW"):
        """Set/query counter trigger sensitivity = LOW|MEDIUM|HIGH (for 0% to 100%)"""
        return self.arg1_with_query( "COUN:SENS", ["LOW", "MEDIUM", "HIGH"], mode, query, debug = debug)
    
    def counter_measure(self, debug = [1000, 0.001, 0.5, 500e-6, 500e-6]):
        """Query counter measure results.
        list [ freq, period, duty_cycle, pos_pulse_period, neg_pulse_period ]
        """
        if self.emulation:
            return debug
        
        values = []
        for item in ["FREQ", "PER", "DCYC", "POSW", "NEGW"]:
            meas = self.get_query( "COUN:" + item )
            values.append( float( meas.strip("%") ) )
        return values
    
    def counter_status(self, mode = None, query = False, debug = "Unknown"):
        """Set counter status = OFF|ON"""
        if query:
            return debug
        self.arg1_with_query( "COUN", ["OFF", "ON"], mode, query = False )
    
    ########
    # Output
    def channel_output_load(self, ch, mode = None, query = False, debug = "50"):
        """Set/query channel output termination load = INF | 50 | any
        ch = 1 | 2
        Load is High Impedance (INF), or 50 ohm, or actually any R.
        Note the instrument output is fixed at 50ohm.
        Return: ohms / Infinity
        """
        return self.arg1_with_query( "OUTP:LOAD" + CH_TEXT[ch], None, mode, query, debug = debug)
    
    def channel_output(self, ch, mode = None, query = False, debug = "0"):
        """Set/query channel output state = OFF | ON | 0 | 1
        ch = 1 | 2
        """
        return self.arg1_with_query( "OUTP" + CH_TEXT[ch], ["0", "1", "OFF", "ON"], mode, query, debug = debug)
    
    ########
    # Signal
    def channel_basic_waveform(self, ch, wtype = None, freq = None, amp = None, offset = None, phase = None, duty = None, sym = None, delay = None, query = False, debug = "SINUSOID"):
        """Channel basic wave setup. A swiss army knife function.
        ch = 1 | 2
        Different waveform types as follows, at least one optional argument must be given.
        wtype='SINUSOID', optional: freq, amp, offset, phase.
        wtype='SQUARE',   optional: freq, amp, offset, phase, duty.
        wtype='RAMP',     optional: freq, amp, offset, phase, sym.
        wtype='PULSE',    optional: freq, amp, offset, duty. delay might not be used
        wtype='NOISE',    optional: amp, offset.
        wtype='ARB',      optional: freq, amp, offset, phase.
        wtype='DC',       optional: offset.
        freq = 1e-6 to 20e6 [Hz]
        amp = 0.004 to 6 or ch2 20 [Vpp]. For NOISE this is variance, 0.4m to 2.222.
        offset = similar to amp [V]. For NOISE this is mean.
        phase = 0 to 360
        duty = 20 to 80 % - Does depend on freq.
        sym = 0 to 100 %
        delay = 0 up
        
        Missing argument are ignored and the instrument will carry on with the previous setting for the missing parameters.
        """
        # Handle query.
        out = "APPL" + CH_TEXT[ch]
        if query:
            return self.get_query(out, debug = debug)
        
        try:
            assembly = BASIC_WAVE_TYPE_ARGS[wtype]
        except:
            raise visacomsbase.VisaComsBaseError("Unknown waveform type.")
        
        # Ensure amplitude unit VPP.
        self.channel_voltage_unit(ch, "VPP")
        
        # Scan all parameters; build wanted, trap unwanted.
        # Note, limits are from Siglent unit.
        for val in assembly:
            if val == "WAVETYPE":
                # Catch 'WAVETYPE' first. Map given wtype to SCPI name.
                self.channel_function_shape(ch, BASIC_WAVE_TYPE_MAP.get(wtype, wtype) )
            elif val == "FREQ" and freq != None:
                self.channel_frequency(ch, freq)
            elif val == "AMP" and amp != None:
                self.channel_voltage_amp(ch, amp)
            elif val == "OFFSET" and offset != None:
                self.channel_voltage_offset(ch, offset)
            elif val == "PHASE" and phase != None:
                self.channel_phase(ch, phase)
                # Align this channel to other.
                self.channel_phase_align(ch)
            elif val == "DUTY" and duty != None:
                if wtype == "SQUARE":
                    self.channel_square_duty(ch, duty)
                else:
                    self.channel_pulse_duty(ch, duty)
            elif val == "SYM" and sym != None:
                self.channel_ramp_symmetry(ch, sym)
            elif val == "DELAY" and delay != None:
                # Convert delay & freq to a phase.
                self.channel_phase(ch, delay * freq * 360)
                # Align this channel to other.
                self.channel_phase_align(ch)
            #else:
            #    raise visacomsbase.VisaComsBaseError("Unknown variable.")
    
    ########
    # Frequency
    def channel_frequency(self, ch, mode = None, query = False, debug = "1000"):
        """Set/query channel frequency = number | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( "FREQ" + CH_TEXT[ch], None, mode, query, debug = debug)
    
    # start all CH1 only.
    # stop
    # center
    # span
    
    ########
    # Function
    def channel_ramp_symmetry(self, ch, mode = None, query = False, debug = "50"):
        """Set/query channel function ramp symmetry = percent | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( "FUNC:RAMP:SYMM" + CH_TEXT[ch], None, mode, query, debug = debug)
    
    def channel_function_shape(self, ch, mode = None, query = False, debug = "SINUSOID"):
        """Set/query channel function (waveform) = SINUSOID | SQUARE | RAMP | PULSE | NOISE | USER | DC
        ch = 1 | 2
        """
        return self.arg1_with_query( "FUNC" + CH_TEXT[ch], ["SINUSOID", "SQUARE", "RAMP", "PULSE", "NOISE", "USER", "DC"], mode, query, debug = debug)
    
    def channel_square_duty(self, ch, mode = None, query = False, debug = "50"):
        """Set/query channel function square duty = percent | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( "FUNC:SQU:DCYC" + CH_TEXT[ch], None, mode, query, debug = debug)
    
    def channel_function_user(self, ch, mode = None, query = False, debug = "user"):
        """Set/query channel user waveform = See manual
        ch = 1 | 2
        """
        return self.arg1_with_query( "FUNC:USER" + CH_TEXT[ch], None, mode, query, debug = debug)
    
    ########
    # Phase
    def channel_phase(self, ch, mode = None, query = False, debug = "0"):
        """Set/query channel initial phase = number -180 to 180 degrees
        ch = 1 | 2
        """
        if not query:
            mode = mode % 360
            if mode > 180:
                mode -= 360
        return self.arg1_with_query( "PHAS" + CH_TEXT[ch], None, mode, query, debug = debug)
    
    def channel_phase_align(self, ch):
        """Set channel phase to alignment with other channel. Ignores given channel.
        ch = 1 | 2
        """
        return self.write( "PHAS:ALIGN" )
    
    ########
    # Pulse
    def channel_pulse_duty(self, ch, mode = None, query = False, debug = "50"):
        """Set/query channel pulse duty cycle = percent | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( "PULS:DCYC" + CH_TEXT[ch], None, mode, query, debug = debug)
    
    def channel_pulse_period(self, ch, mode = None, query = False, debug = "50"):
        """Set/query channel pulse duty cycle = percent | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( "PULS:PER" + CH_TEXT[ch], None, mode, query, debug = debug)
    
    def channel_pulse_width(self, ch, mode = None, query = False, debug = "500e-6"):
        """Set/query channel pulse width = number | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( "PULS:WIDT" + CH_TEXT[ch], None, mode, query, debug = debug)
    
    ########
    # Voltage
    def channel_voltage_amp(self, ch, mode = None, query = False, debug = "1"):
        """Set/query channel voltage amplitude = number (VPP) | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( "VOLT" + CH_TEXT[ch], None, mode, query, debug = debug)
    
    # high
    # low
    
    def channel_voltage_offset(self, ch, mode = None, query = False, debug = "0"):
        """Set/query channel voltage offset = number | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( "VOLT:OFFS" + CH_TEXT[ch], None, mode, query, debug = debug)
    
    def channel_voltage_unit(self, ch, mode = None, query = False, debug = "VPP"):
        """Set/query channel voltage unit = VPP | VRMS | DBM
        ch = 1 | 2
        """
        return self.arg1_with_query( "VOLT:UNIT" + CH_TEXT[ch], ["VPP", "VRMS", "DBM"], mode, query, debug = debug)
    
    ########
    # System
