#!/usr/bin/env python
"""Arbitrary waveform generator interface - ethernet, GPIB, RS232.

Rigol DG1000Z series arbitrary waveform generator.

Validated with instrument.

Author: Ian Watson (igw)
Date: 11 Dec 2014
"""

#import time

import visacomsbase


################################################################################
# Channels
CH1 = 1
CH2 = 2

# Basic waveform type arguments.
BASIC_WAVE_TYPE_ARGS = {
    "SINUSOID": ["WAVETYPE", "FREQ", "AMP", "OFFSET", "PHASE"],
      "SQUARE": ["WAVETYPE", "FREQ", "AMP", "OFFSET", "PHASE", "DUTY"],
        "RAMP": ["WAVETYPE", "FREQ", "AMP", "OFFSET", "PHASE", "SYM"],
       "PULSE": ["WAVETYPE", "FREQ", "AMP", "OFFSET", "DELAY", "DUTY"],
       "NOISE": ["WAVETYPE", "AMP", "OFFSET"],
         "ARB": ["WAVETYPE", "FREQ", "AMP", "OFFSET", "PHASE"],
          "DC": ["WAVETYPE", "OFFSET"]
}

# Map wave type given to SCPI command, either shorthand or a different name.
BASIC_WAVE_TYPE_MAP = { "SINUSOID" : "SIN", "SQUARE" : "SQU", "ARB" : "USER" }


class AWG_DG1000Z(visacomsbase.VisaComsBase):
    """Interface to Rigol DG1000Z series arbitrary waveform generator."""
    
    # Default resource name.
    resource_name = "USBInstDG1000Z"
    # Dict of arguments for visa interface.
    args_visa = { "read_termination" : "\n" }
    # Dict of arguments for visacoms module.
    # Instrument seems fast. Use delay 0.05s for a bit of margin.
    # Does this instrument respond to OPC?
    args_visacoms = {
        "what_am_i" : "I might be a Rigol DG1000Z series arbitrary waveform generator.",
        "prompt" : "AWG_DG1000Z",
        "read_term_chars_strip" : '\n\r',
        "write_delay" : 0.05 #, "read_delay" : 0.05
    }
    
    # Inherit __init__
    
    # Channels.
    CH1 = CH1
    CH2 = CH2
    # All channels.
    CHANS = ( CH1, CH2 )
    # Max/Min levels into 50ohm.
    MAX_50R_VPK = 5.0
    MIN_50R_VPK = 0.001
    
    ############################################################################
    # Counter
    def counter_coupling(self, mode = None, query = False, debug = "AC"):
        """Set/query counter coupling = AC | DC"""
        return self.arg1_with_query( ":COUN:COUP", ["AC", "DC"], mode, query, debug = debug)
    
    def counter_gatetime(self, mode = None, query = False, debug = "USER4"):
        """Set/query counter coupling = USER1|USER2|USER3|USER4|USER5|USER6
        The are:- 1.310ms 10.48ms 166.7ms 1.342s 10.73s >10s
        """
        return self.arg1_with_query( ":COUN:GATE", ["USER1","USER2", "USER3", "USER4", "USER5", "USER6"], mode, query, debug = debug)
    
    def counter_hf_filter(self, mode = None, query = False, debug = "0"):
        """Set/query counter high frequency filter = OFF | ON | 0 | 1"""
        return self.arg1_with_query( ":COUN:HF", ["0", "1", "OFF", "ON"], mode, query, debug = debug)
    
    def counter_trigger_level(self, mode = None, query = False, debug = "0"):
        """Set/query counter trigger level = -2.5V to 2.5V"""
        return self.arg1_with_query( ":COUN:LEVE", None, mode, query, debug = debug)
    
    def counter_trigger_sensitivity(self, mode = None, query = False, debug = "50"):
        """Set/query counter trigger sensitivity = 0% to 100%"""
        return self.arg1_with_query( ":COUN:SENS", None, mode, query, debug = debug)
    
    def counter_measure(self, debug = [1000, 0.001, 0.5, 500e-6, 500e-6]):
        """Query counter measure results.
        list [ freq, period, duty_cycle, pos_pulse_period, neg_pulse_period ]
        """
        if self.emulation:
            return debug
        
        mess = self.get_query( ":COUN:MEAS", debug = debug)
        return [ float(x) for x in mess.split(",") ]
    
    def counter_status(self, mode = None, query = False, debug = "0"):
        """Set/query counter status = ON|1|OFF|0|RUN|STOP|SINGLE"""
        return self.arg1_with_query( ":COUN:STAT", ["ON", "OFF", "RUN", "STOP", "SINGLE"], mode, query, debug = debug)
    
    def counter_status_wait(self, till = 'STOP'):
        """Wait till counter status changes to given. Normally used to wait for 'STOP'."""
        while not self.emulation:
            self.sleep_em(0.1)
            mess = self.counter_status(query = True, debug = till)
            #print mess
            if mess == till:
                break
    
    def counter_stats_clear(self, mode = None, query = False, debug = "0"):
        """Statistic clear results"""
        return self.write( ":COUN:STATI:CLEA" )
    
    def counter_stats_display(self, mode = None, query = False, debug = "DIGITAL"):
        """Set/query counter statistic display format = DIGITAL | CURVE"""
        return self.arg1_with_query( ":COUN:STATI:DISP", ["DIGITAL", "CURVE"], mode, query, debug = debug)
    
    def counter_stats_status(self, mode = None, query = False, debug = "0"):
        """Set/query counter statistic status = OFF | ON | 0 | 1"""
        return self.arg1_with_query( ":COUN:STATI:STAT", ["0", "1", "OFF", "ON"], mode, query, debug = debug)
    
    ########
    # Output
    def channel_output_load(self, ch, mode = None, query = False, debug = "50"):
        """Set/query channel output termination load = INF | 50 | any
        ch = 1 | 2
        Load is High Impedance (INF), or 50 ohm, or actually any R.
        Note the instrument output is fixed at 50ohm.
        Return: ohms
        """
        return self.arg1_with_query( ":OUTP" + str(ch) + ":LOAD", None, mode, query, debug = debug)
    
    def channel_output(self, ch, mode = None, query = False, debug = "0"):
        """Set/query channel output state = OFF | ON | 0 | 1
        ch = 1 | 2
        """
        return self.arg1_with_query( ":OUTP" + str(ch) + ":STAT", ["0", "1", "OFF", "ON"], mode, query, debug = debug)
    
    ########
    # Signal
    def channel_basic_waveform(self, ch, wtype = None, freq = None, amp = None, offset = None, phase = None, duty = None, sym = None, delay = None, query = False, debug = "SINUSOID"):
        """Channel basic wave setup. A swiss army knife function.
        ch = 1 | 2
        Different waveform types as follows, at least one optional argument must be given.
        wtype='SINUSOID', optional: freq, amp, offset, phase.
        wtype='SQUARE',   optional: freq, amp, offset, phase, duty.
        wtype='RAMP',     optional: freq, amp, offset, phase, sym.
        wtype='PULSE',    optional: freq, amp, offset, duty. delay might not be used
        wtype='NOISE',    optional: amp, offset.
        wtype='ARB',      optional: freq, amp, offset
        wtype='DC',       optional: offset.
        freq = 1e-6 to 20e6 [Hz]
        amp = 0.004 to 6 or ch2 20 [Vpp]. For NOISE this is variance.
        offset = similar amp. For NOISE this is mean.
        phase = 0 to 360
        duty = 20 to 80 % - Does depend on freq.
        sym = 0 to 100 %
        delay = 0 up
        
        Missing argument are ignored and the instrument will carry on with the previous setting for the missing parameters.
        """
        # Handle query.
        if query:
            return self.channel_function_shape(ch, query = query, debug = debug)
        
        try:
            assembly = BASIC_WAVE_TYPE_ARGS[wtype]
        except:
            raise visacomsbase.VisaComsBaseError("Unknown waveform type.")
        
        # Ensure amplitude unit VPP.
        self.channel_voltage_unit(ch, "VPP")
        
        # Scan all parameters; build wanted, trap unwanted.
        # Note, limits are from Siglent unit.
        for val in assembly:
            if val == "WAVETYPE":
                # Catch 'WAVETYPE' first. Map given wtype to SCPI name.
                self.channel_function_shape(ch, BASIC_WAVE_TYPE_MAP.get(wtype, wtype) )
            elif val == "FREQ" and freq != None:
                self.channel_frequency(ch, freq)
            elif val == "AMP" and amp != None:
                self.channel_voltage_amp(ch, amp)
            elif val == "OFFSET" and offset != None:
                self.channel_voltage_offset(ch, offset)
            elif val == "PHASE" and phase != None:
                self.channel_phase(ch, phase)
                # Align this channel to other.
                self.channel_phase_align(ch)
            elif val == "DUTY" and duty != None:
                if wtype == "SQUARE":
                    self.channel_square_duty(ch, duty)
                else:
                    self.channel_pulse_duty(ch, duty)
            elif val == "SYM" and sym != None:
                self.channel_ramp_symmetry(ch, sym)
            elif val == "DELAY" and delay != None:
                # Convert delay & freq to a phase.
                self.channel_phase(ch, delay * freq * 360)
                # Align this channel to other.
                self.channel_phase_align(ch)
            #else:
            #    raise visacomsbase.VisaComsBaseError("Unknown variable.")
    
    ########
    # Frequency
    def channel_frequency(self, ch, mode = None, query = False, debug = "1000"):
        """Set/query channel frequency (fixed) = number | MIN | MAX
        For basic waveforms.
        ch = 1 | 2
        """
        return self.arg1_with_query( ":SOUR" + str(ch) + ":FREQ:FIX", None, mode, query, debug = debug)
    
    # start all CH1 only.
    # stop
    # center
    # span
    
    ########
    # Function
    def channel_ramp_symmetry(self, ch, mode = None, query = False, debug = "50"):
        """Set/query channel function ramp symmetry = percent | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( ":SOUR" + str(ch) + ":FUNC:RAMP:SYMM", None, mode, query, debug = debug)
    
    def channel_function_shape(self, ch, mode = None, query = False, debug = "SINUSOID"):
        """Set/query channel function (waveform shape) = SINUSOID | SQUARE | RAMP | PULSE | NOISE | USER | DC | and lots more.
        ch = 1 | 2
        """
        return self.arg1_with_query( ":SOUR" + str(ch) + ":FUNC:SHAP", None, BASIC_WAVE_TYPE_MAP.get(mode, mode), query, debug = debug)
    
    def channel_square_duty(self, ch, mode = None, query = False, debug = "50"):
        """Set/query channel function square duty = percent | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( ":SOUR" + str(ch) + ":FUNC:SQU:DCYC", None, mode, query, debug = debug)
    
    ########
    # Phase
    def channel_phase(self, ch, mode = None, query = False, debug = "0"):
        """Set/query channel initial phase = number 0 to 360 degrees
        ch = 1 | 2
        """
        if not query:
            mode = mode % 360
        return self.arg1_with_query( ":SOUR" + str(ch) + ":PHAS:ADJ", None, mode, query, debug = debug)
    
    def channel_phase_align(self, ch):
        """Set channel phase to alignment with other channel.
        ch = 1 | 2
        """
        return self.write( ":SOUR" + str(ch) + ":PHAS:SYNC" )
    
    ########
    # Pulse
    def channel_pulse_duty(self, ch, mode = None, query = False, debug = "50"):
        """Set/query channel pulse duty cycle = percent | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( ":SOUR" + str(ch) + ":PULS:DCYC", None, mode, query, debug = debug)
    
    def channel_pulse_period(self, ch, mode = None, query = False, debug = "1e-3"):
        """Set/query channel pulse period = number | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( ":SOUR" + str(ch) + ":FUNC:PULS:PER", None, mode, query, debug = debug)
    
    def channel_pulse_width(self, ch, mode = None, query = False, debug = "500e-6"):
        """Set/query channel pulse width = number | MIN | MAX
        ch = 1 | 2
        Also :SOUR<n>:PULS:WIDT
        """
        return self.arg1_with_query( ":SOUR" + str(ch) + ":FUNC:PULS:WIDT", None, mode, query, debug = debug)
    
    ########
    # Voltage
    def channel_voltage_amp(self, ch, mode = None, query = False, debug = "1"):
        """Set/query channel voltage amplitude = number (VPP) | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( ":SOUR" + str(ch) + ":VOLT:LEV:IMM:AMPL", None, mode, query, debug = debug)
        #return self.arg1_with_query( ":SOUR" + str(ch) + ":VOLT:LEV", None, mode, query, debug = debug)
    
    # high
    # low
    
    def channel_voltage_offset(self, ch, mode = None, query = False, debug = "0"):
        """Set/query channel voltage offset = number | MIN | MAX
        ch = 1 | 2
        """
        return self.arg1_with_query( ":SOUR" + str(ch) + ":VOLT:LEV:IMM:OFFS", None, mode, query, debug = debug)
    
    def channel_voltage_unit(self, ch, mode = None, query = False, debug = "VPP"):
        """Set/query channel voltage unit = VPP | VRMS | DBM
        ch = 1 | 2
        """
        return self.arg1_with_query( ":SOUR" + str(ch) + ":VOLT:UNIT", ["VPP", "VRMS", "DBM"], mode, query, debug = debug)
    
    ########
    # System
