#!/usr/bin/env python
"""Digital Oscilloscope interface - USB.

Common parts.
Author: Ian Watson (igw)
Date: 5 Jun 2015
"""

#import sys
#import os
import visacomsbase


################################################################################
class DS_Common(visacomsbase.VisaComsBase):
    """Common parts for interface to digital oscilloscope."""
    
    ############################################################################
    def channel_scale_preferred(self, chan, mode = None, query = False, debug = "1"):
        """Set/query channel scale (per division) to preferred values = number
        OVER LORD FUNCTION.
        This is a multiple of 1, 2, 5 steps.
        FIXME: Write this.
        """
        if not query:
            mode = 0
        return self.channel_scale(chan, mode, query, debug = debug)
    
    ############################################################################
    def timebase_screen_cycles(self, freq, cps = 3):
        """Determine timebase_scale for given number of sinusoid cycles per screen, typically 3.
        Additionally limit time scale to given time per division maximum, usually 10m or 1m.
        This ensures the scope does fast scans, not rolling waveform.
        Also the Hameg has slow wave data fetch for time scale > 1ms. Why? No idea.
        - self.time_div_max - Attached to scope class master or instance override.
        """
        return min(1.0 / freq * cps / self.DIVISIONSX, self.time_div_max)
    
    # Auto measurement add ons.
    def measure_do_amplitude_fit(self, chan, target = 0.8, munit = 6, debug = 1):
        """Measure a signal and adjust y-scale for best amplitude fit.
        VPP used for signal fit.
        Assumes the signal is AC coupled and offset=0, i.e., centred.
        Around 2-3 cycle of a repetitive signal provide best measurement.
        This simple algorithm needs time to settle if AC coupling in play.
        TODO: Issue: The Rigol scope has a relay switch between 0.2-0.5 V/div with x1 probe. 2-5V x10 probe. This causes some disturbance. A delay to handle this would solve an issue.
        - chan - CHAN1, etc.
        - munit - If used, different unit from default.
        - target - Target amplitude as fraction of full scale.
        Return, final value of channel_scale().
        """
        if self.emulation:
            return debug
        time_scale = self.timebase_scale(query = True)
        y_scale = self.channel_scale(chan, query = True)
        y_idx = self.idx2scale(self.YDIVSCALES, y_scale)
        probe = self.channel_probe(chan, query = True)
        y_idx_min, y_idx_max = self.YDIVSCALEBYPROBE[probe]
        # Count fail safe.
        count = 0
        while count < len(self.YDIVSCALES):
            y_max = self.YDIVSCALES[y_idx] * self.DIVISIONSY
            y_max_up = self.YDIVSCALES[ max(y_idx_min, y_idx-1) ] * self.DIVISIONSY
            #print "YYYY", y_max, y_max_up, y_idx_min, y_idx_max, max(y_idx_min, y_idx-1)
            vpp = self.measure_item("VPP", chan, munit = munit, or_repeat = 0, query = True)
            # Target 80% with 5% hist. Check if to big - reduce level, inc scale. To small, dec scale, ensure x2.5 still fits.
            if vpp > y_max * target * 1.05 and y_idx < y_idx_max:
                y_idx += 1
            elif vpp < y_max_up * target * 0.95 and y_idx > y_idx_min:
                y_idx -= 1
            else:
                break
            self.channel_scale(chan, self.YDIVSCALES[y_idx])
            # FIXME: Add function to change this time depending upon actions.
            # Hameg switches between 0.1-0.2V/div on x1. Between 1-2V/div on x10.
            # Hameg settles quickly on change so delay can probably go.
            # Check when rigol's do this.
            self.sleep_em( 0.5 + time_scale * self.DIVISIONSX )
            count += 1
        return self.YDIVSCALES[y_idx]
    
    def idx2scale(self, values, value):
        """Index to value in list which matches or is below.
        So 2 in [1,2,5] is index 1, i.e. 2 and matches.
        So 2.5 in [1,2,5] is index 2, i.e. 5 and is below.
        """
        idx = 0
        # Scan till equal or above.
        #while abs( value - values[y_idx] ) > 0.0001:
        while value > values[idx] + 1e-6:
            idx += 1
        return idx
    
    ############################################################################
#     def measure_play(self, ch):
#         import math
#         import numpy
#         import matplotlib.pyplot as plt
#         
#         tsamp = 100e-6
#         fsamp =  1.0 / tsamp
#     
#         tin = 8 * tsamp
#         fin = 1.0 / tin
#     
#         tlen = 4 * tin
#         flen = 1.0 / tlen
#     
#         t_run = [ tsamp * n for n in range(0, int(tlen / tsamp)) ]
#     
#         v1 = [ math.sin( 2 * math.pi * fin * t ) for t in t_run ]
#         ta = tsamp * numpy.arange(0, int(tlen / tsamp))
#         v2 = numpy.sin( 2 * math.pi * fin * ta )
#         plt.plot(v2)
#         #plt.ion()
#         plt.show()
#     
#         # FFT with amplitude scale correction, i.e, a true DFT. FFT returns 'n' points, where is pads if n not power of 2.
#         L = len(v2)
#         spec = numpy.fft(v2) / L
#         spec = numpy.absolute(spec)
#         spec = numpy.fft.fftshift(spec)
#         # Frequency bin resolution and frequency axis.
#         fbin = 1 / ( L * tsamp )
#         ftmp = fbin * numpy.arange(-L/2, L/2)
#     
#         plt.plot(ftmp, spec)
#         #plt.ioff()
#         plt.show()
#         
#         v3 = 1
