#!/usr/bin/env python
"""Digital Oscilloscope interface - USB.

Rigol DS1000E, D series digital oscilloscopes.
Validated with instrument.

Note this is not a full implementation. Just the necessary and basic's functions.

Author: Ian Watson (igw)
Date: 3 Nov 2015
"""

import numpy

import visacomsbase
import DS_Common
import dataholder


################################################################################
# Channels by name.
CHANOFF = "OFF"
CHAN1 = "CHAN1"
CHAN2 = "CHAN2"
CHANMATH = "MATH"
# Channel names. By index, 0 is OFF, Chan 1-2. 3 is MATH.
CHANNAMES = ( CHANOFF, CHAN1, CHAN2, CHANMATH )

class DS_DS1000E(DS_Common.DS_Common):
    """Interface to digital oscilloscope."""
    
    # Default resource name.
    resource_name = "USBInstDS1000E"
    # Dict of arguments for visa interface.
    # Takes long time to read 1Mpt long memory, so increase timeout. No "read_termination" : "\n" required.
    args_visa = { "timeout" : 15000, "chunk_size" : 1100000 }
    # Dict of arguments for visacoms module.
    # A delay after each write. Give time for instrument to process. Initially used 0.5. Was mainly for measure_do_amplitude_fit, but have added delay into it.
    # Delay before read. 0 worked.
    args_visacoms = {
        "what_am_i" : "I might be a Rigol DS1000E,D series digital oscilloscope.",
        "prompt" : "DS_DS1000E",
        "write_delay" : 0.3
        #"read_term_chars_strip" : '\n\r'
    }
    # This instrument does not respond to OPC?
    
    # Inherited __init__
    
    # Channels.
    CHAN1 = CHAN1
    CHAN2 = CHAN2
    # All channels.
    CHANS = ( CHAN1, CHAN2 )
    # Device parameters.
    DIVISIONSY = 8
    DIVISIONSX = 12
    # All available y-scales.
    YDIVSCALES = [0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5, 10, 20, 50, 100]
    # Range of scales (inclusive) for each probe attenuation. Index to YDIVSCALES.
    YDIVSCALEBYPROBE = {"1" : (0, 11), "5" : (2, 13), "10" : (3, 14)}
    #
    time_div_max = 0.01
    
    ############################################################################
    # Acquisition commands.
    def system_run(self):
        """Starts continuous acquisitions until STOP or SING."""
        # A RUN command cause this scope to start a run. If in SINGLE mode, it does only 1 run, then stops. So we must switch to AUTO mode to do what RUN does with most scopes.
        self.trigger_sweep_mode("AUTO")
        self.write(":RUN")
    
    def system_stop(self):
        """Stops the running acquisition. MAY TRUNCATE DATA CAPTURE. Use system_single.
        OK this is data safe on Rigol, but not on Hameg.
        """
        self.write(":STOP")
    
    def system_single(self):
        """Starts (it triggers) a defined number of acquisition cycles."""
        self.trigger_sweep_mode("SING")
        #self.trigger_force()
        #self.write(":RUN")
    
    def ieee_opc_get(self):
        """IEEE *OPC?
        Returns 1 when previous command has completed.
        """
        return "1"
    
    ########
    # :ACQuire
    def acquire_memory_depth(self, mode = None, query = False, debug = "NORM"):
        """Set/query acquire memory depth = NORMal|LONG
        Only works when using :MEMDepth
        """
        mess = self.arg1_with_query(":ACQ:MEMDepth", ["NORM", "LONG"], mode, query, debug = debug)
        return mess
    
    def acquire_type(self, mode = None, query = False, debug = "NORM"):
        """Set/query acquire type = NORMal|AVERage|PEAKdetect
        Oddly, returns 'Peak Detect' not PEAKDETECT.
        """
        mess = self.arg1_with_query(":ACQ:TYPE", ["NORM", "AVER", "PEAK"], mode, query, debug = debug)
        return mess
    
    def acquire_aver_count(self, mode = None, query = False, debug = "16"):
        """Set/query average count. 2, 4, 8, to 1024, i.e., 2^N, N = 1-10"""
        mess = self.arg1_with_query("ACQ:AVER", None, mode, query, debug = debug)
        if query:
            return int(mess)
    
    def acquire_aver_wait_till_done(self):
        """Query, waits for average count to complete."""
        if self.acquire_type(query = True) == "AVER":
            ts = self.timebase_scale(query = True)
            ave_num = self.acquire_aver_count(query = True)
            self.sleep_em( ave_num * (ts + 0.001) * self.DIVISIONSX )
        
        return 1
    
    def acquire_sampling_rate(self, debug = 100e6):
        """Query acquire sampling rate query."""
        mess = self.get_query(":ACQ:SAMP", debug = debug)
        return float(mess)
    
    ########
    # :CHANnel
    SCPI_MAP_BANDWIDTH_LIMIT = visacomsbase.bidirectional_name_map( {"20M" : "ON"} )
    def channel_bandwidth_limit(self, chan, mode = None, query = False, debug = "OFF"):
        """Set/query channel bandwidth limit = 20M|OFF - actual ON|OFF
        - chan - CHAN1|CHAN2
        """
        mode = self.SCPI_MAP_BANDWIDTH_LIMIT.get(mode, mode)
        mess = self.arg1_with_query(":" + chan + ":BWL", ["ON", "OFF"], mode, query, debug = debug)
        return self.SCPI_MAP_BANDWIDTH_LIMIT.get(mess, mess)
    
    def channel_coupling(self, chan, mode = None, query = False, debug = "DC"):
        """Set/query channel coupling = DC|AC|GND
        - chan - CHAN1|CHAN2
        Wait for AC to settle.
        """
        mess = self.arg1_with_query(":" + chan + ":COUP", ["DC", "AC", "GND"], mode, query, debug = debug)
        if not query and mode == "AC":
            self.sleep_em(1.2)
        return mess
    
    def channel_display(self, chan, mode = None, query = False, debug = "OFF"):
        """Set/query channel display = ON|OFF
        - chan - CHAN1|CHAN2
        """
        mess = self.arg1_with_query(":" + chan + ":DISP", ["ON", "OFF"], mode, query, debug = debug)
        return mess
    
    def channel_filter(self, chan, mode = None, query = False, debug = "OFF"):
        """Set/query channel filter = ON|OFF
        - chan - CHAN1|CHAN2
        Of little use, required by _waveform_get_data to ensure it works.
        """
        mess = self.arg1_with_query(":" + chan + ":FILT", ["ON", "OFF"], mode, query, debug = debug)
        return mess
    
    def channel_memory_depth(self, chan, debug = "1e6"):
        """Query memory depth. Say 8192 (8Kpts), 16384, 512K, 1M.
        - chan - CHAN1|CHAN2
        """
        mess = self.get_query(":" + chan + ":MEMD", debug = debug)
        return int(mess)
    
    def channel_offset(self, chan, mode = None, query = False, debug = "0"):
        """Set/query channel offset = number
        - chan - CHAN1|CHAN2
        """
        mess = self.arg1_with_query(":" + chan + ":OFFS", None, mode, query, debug = debug)
        if query:
            return float(mess)
    
    def channel_offset_position(self, chan, mode = None, query = False, debug = "0"):
        """Dummy function. Set/query channel offset position (in divisions) = -5 <= divisions <= 5
        - chan - CHAN1|CHAN2
        """
        scale = self.channel_scale(chan, query = True)
        if query:
            mess = self.channel_offset(chan, mode, query, debug = debug) / scale
            return mess
        mode = mode * scale
        self.channel_offset(chan, mode, query)
    
    def channel_probe(self, chan, mode = None, query = False, debug = "10"):
        """Set/query channel probe (attenuation factor) = 1|5|10|50|100|500|1000
        - chan - CHAN1|CHAN2
        """
        mess = self.arg1_with_query(":" + chan + ":PROB", ["1", "5", "10", "50", "100", "500", "1000"], str(mode), query, debug = debug)
        if query:
            # Returns str "1.000e0" format! Convert to tidy sting.
            return str( int( float(mess) ) )
    
    def channel_probe_type(self, chan):
        """Dummy function. Query channel probe type = NONE | ACTive | PASSive
        - chan - CHAN1|CHAN2
        """
        return "NONE"
    
    def channel_scale(self, chan, mode = None, query = False, debug = "1"):
        """Set/query channel scale (per division) = number
        x1 is 2mV to 10V. x10 is 20mV to 100V. Usual 1, 2, 5, 10 scale.
        However it will take any number, as the venier has not affect on this.
        - chan - CHAN1|CHAN2
        """
        mess = self.arg1_with_query(":" + chan + ":SCAL", None, mode, query, debug = debug)
        if query:
            return float(mess)
    
    def channel_venier(self, chan, mode = None, query = False, debug = "OFF"):
        """Set/query channel scale vernier = ON|OFF
        - chan - CHAN1|CHAN2
        """
        mess = self.arg1_with_query(":" + chan + ":VERN", ["ON", "OFF"], mode, query, debug = debug)
        return mess
    
    ########
    # :DISPlay
    def display_grid(self, mode = None, query = False, debug = "FULL"):
        """Set/query display grid = FULL|HALF|NONE"""
        mess = self.arg1_with_query(":DISP:GRID", ["FULL", "HALF", "NONE"], mode, query, debug = debug)
        return mess
    
    ########
    # :MATH
    def math_display(self, mode = None, query = False, debug = "OFF"):
        """Set/query math display = ON|OFF"""
        mess = self.arg1_with_query(":MATH:DISP", ["ON", "OFF"], mode, query, debug = debug)
        return mess
    
    def math_operate(self, mode = None, query = False, debug = "A+B"):
        """Set/query math operate = A+B|A-B|AB|FFT"""
        mess = self.arg1_with_query(":MATH:OPER", ["A+B", "A-B", "AB", "FFT"], mode, query, debug = debug)
        return mess
    
    def fft_display(self, mode = None, query = False, debug = "OFF"):
        """Set/query FFT display = ON|OFF"""
        mess = self.arg1_with_query(":FFT:DISP", ["ON", "OFF"], mode, query, debug = debug)
        return mess
    
    ########
    # :MEAS
    def measure_source(self, mode = None, query = False, debug = "CH1"):
        """Set/query measurement source = CHAN1|CHAN2
        Return CH1, CH2"""
        mess = self.arg1_with_query(":MEAS:SOUR", ["CHAN1", "CHAN2"], mode, query, debug = debug)
        return mess
    
    def measure_total(self, mode = None, query = False, debug = "OFF"):
        """Set/query measurement total = ON|OFF"""
        mess = self.arg1_with_query(":MEAS:TOT", ["ON", "OFF"], mode, query, debug = debug)
        return mess
    
    MEASURE_ITEM_MAP = visacomsbase.bidirectional_name_map( {"VAVG" : "VAV"} )
    def measure_item(self, mode, chan = None, chanref = None, munit = 6, or_repeat = 8, query = True, debug = 1):
        """Get measurement from channel of type taken from, Rigol DS1000Z names,
        MAX | VMIN | VPP | VTOP | VBASe | VAMP | VAVG | VRMS | OVERshoot | PREShoot | MARea | MPAReaPERiod | FREQuency | RTIMe | FTIMe | PWIDth | NWIDth | PDUTy | NDUTy | RDELay | FDELay | RPHase | FPHase
        These are mapped to the DS1000E names,
        VPP | VMAX | VMIN | VAMP | VTOP | VBASe | VAVerage | VRMS | OVERshoot | PREShoot | FREQ | RISetime | FALLtime | PWIDth | NWIDth | PDUTycycle | NDUTycycle | PDELay | NDELay | TOT
        
        Returns '9.9e37' if over level.
        
        - chan - str of source, usually channel, say CHAN1.
        - chanref - 2nd channel reference.
        - munit - Measurement unit if appropriate. Not used, for compat.
        - or_repeat - Overrange/overload repeat measurement count.
        - query - Not used. For DS1000Z compatibility.
        """
        if self.emulation:
            return debug
        mode = self.MEASURE_ITEM_MAP.get(mode, mode)
        out = ":MEAS:" + mode + "?"
        if chan is not None:
            out += " " + chan
        
        # Do measurement.
        mess = self._measure_item_cmd(out)
        
        if mess >= 1e37 and or_repeat > 0:
            print(">> Measure over range. Retry %d times" % ( or_repeat ), end = '')
            for count_n in range(or_repeat):
                mess = self._measure_item_cmd(out)
                print(", %g" % (mess), end = '')
                if mess < 1e37:
                    break
            print(".")
            if mess >= 1e37:
                print(">> Retry unsuccessful.")
        
        return mess
    
    def _measure_item_cmd(self, out):
        """The measure item command."""
        self.write(out)
        # I think a little wait can only help here.
        #self.sleep_em(0.2)
        mess = self.read()
        # Returns '*****' if over level.
        if mess[0] == "*":
            mess = 9.9e37
        
        return float(mess)
    
    ########
    # :SYSTem
    def beep_make(self):
        """Make a beep."""
        self.write(":BEEP:ACT")
    
    ########
    # :TIMebase
    def timebase_mode(self, mode = None, query = False, debug = "MAIN"):
        """Set/query timebase mode = MAIN|DELayed"""
        mess = self.arg1_with_query(":TIM:MODE", ["MAIN", "DEL"], mode, query, debug = debug)
        return mess
    
    def timebase_offset(self, mode = None, query = False, debug = "0"):
        """Set/query timebase offset = number"""
        mess = self.arg1_with_query(":TIM:OFFS", None, mode, query, debug = debug)
        if query:
            return float(mess)
    
    def timebase_scale(self, mode = None, query = False, debug = "1e-6"):
        """Set/query timebase scale (per division) = number of 2ns - 50s
        Time will be rounded to nearest 1,2,5 notation by scope.
        BUG: Scope only takes 0.00000002 format.
        """
        if mode:
            mode = "%0.9f" % mode
        mess = self.arg1_with_query(":TIM:SCAL", None, mode, query, debug = debug)
        if query:
            return float(mess)
    
    ########
    # :TRIGger
    def trigger_type(self, mode = None, query = False, debug = "EDGE"):
        """Set/query trigger mode = EDGE + others
        ONLY implementing EDGE.
        """
        mess = self.arg1_with_query(":TRIG:MODE", ["EDGE"], mode, query, debug = debug)
        if not query:
            self._trig_type = mode
        return mess
    
    def trigger_edge_coupling(self, mode = None, query = False, debug = "DC"):
        """Set/query trigger coupling = DC|AC|LF|HF"""
        mess = self.arg1_with_query(":TRIG:" + self._trig_type + ":COUP", ["DC", "AC", "LF", "HF"], mode, query, debug = debug)
        return mess
    
    def trigger_status(self, debug = "AUTO"):
        """Query trigger status. Return T'D, WAIT, RUN, AUTO, STOP."""
        mess = self.get_query(":TRIG:STAT", debug = debug)
        return mess
    
    def trigger_sweep_mode(self, mode = None, query = False, debug = "AUTO"):
        """Set/query trigger sweep mode = AUTO|NORMal|SINGle
        Use system_single() for a single sweep.
        Valid data will be available after single sweep is triggered or force trigger given.
        To resume continuous capture, change back to AUTO|NORM, or system_run().
        """
        mess = self.arg1_with_query(":TRIG:" + self._trig_type + ":SWE", ["AUTO", "NORM", "SING"], mode, query, debug = debug)
        return mess
    
    # For EDGE
    def trigger_alltype_level(self, mode = None, query = False, debug = "0"):
        """Set/query trigger oftype level = voltage level number"""
        mess = self.arg1_with_query(":TRIG:" + self._trig_type + ":LEV", None, mode, query, debug = debug)
        if query:
            return float(mess)
    
    def trigger_alltype_source(self, mode = None, query = False, debug = "CH1"):
        """Set/query trigger oftype source = CHAN1|CHAN2|EXT|ACLINE
        Return CH1, CH2, EXT, ACLINE"""
        #mess = self._quick_trigger_and_query(":SOUR", ["CHAN1", "CHAN2", "EXT", "ACLINE"], mode, query, debug = debug)
        mess = self.arg1_with_query(":TRIG:" + self._trig_type + ":SOUR", ["CHAN1", "CHAN2", "EXT", "ACLINE"], mode, query, debug = debug)
        return mess
    
    def trigger_force(self):
        """Trigger force trigger. For NORM|SING mode."""
        self.write(":FORC")
    
    ########
    # :WAVeform
    def waveform_mode(self, mode = None, query = False, debug = "MAX"):
        """Set/query waveform point mode = NORMal|MAX|RAW
        NB This does not actually do anything. Data amount depends on RUN/STOP and current channel config.
        See waveform_get_data().
        """
        mess = self.arg1_with_query(":WAV:POIN:MODE", ["NORM", "MAX", "RAW"], mode, query, debug = debug)
        return mess
    
    def waveform_get_data(self, chan, points = None):
        """Get waveform data from source, chan.
        - chan - CHAN1|CHAN2|MATH|FFT
        - points - Number of points to return, or maximum available.
        In RUN, returns only 600 points.
        Use STOP, call, RUN to get full 8/16k NORM memory or 512/1024k LONG memory.
        """
        # Start a preamble data structure. Modify to suit data below.
        wavedata = self.waveform_preamble(chan)
        # Get actual data.
        nums = self._waveform_get_data(chan)
        # Check data length.
        if len(nums) not in [600, 2**13, 2**14, 2**19, 2**20, 2**19-10, 2**20-10]:
            raise visacomsbase.VisaComsBaseError("Did not get expected number of points")
        # Bug 512K * 1M pts are 10 short. Append a repeat to last 10.
        if len(nums) in [2**19-10, 2**20-10]:
            a = numpy.repeat(nums[-1], 10)
            nums = numpy.append(nums, a)
        
        # Correct x time sample rate if screen data provided. Screen provides only 600 points.
        if len(nums) < 610:
            wavedata.xincrement = wavedata.xscale / 50.0
        
        # Short cut length if requested
        if points is not None:
            nums = nums[:points]
        # Change y-axis data.
        wavedata.data_y = (240 - nums) * wavedata.yincrement - ( wavedata.yreference + wavedata.yscale * 4.6 )
        # Ensure correct length.
        wavedata.points = len(wavedata.data_y)
        # Create x-axis data.
        wavedata.data_x = numpy.arange(0, wavedata.points) * wavedata.xincrement
        return wavedata
    
    def _waveform_get_data(self, chan):
        """Get waveform data from, CHAN1|CHAN2|MATH|FFT
        - chan - CHAN1|CHAN2|MATH|FFT
        RUN mode returns 600 pts. STOP mode returns memory depth.
        Points mode NORM RAW   RAW
        Memory           Norm  Long 
        MATH        600  600   600
        FFT         512  512   512
        CHx         600  8192  524288
        1/2 CH      600  16384 1048576
        BUG Known: The first call to this will only return 600 points.
        Raw byte has value 15 to 240. Top LCD is 25, middle 125, bottom 225.
        So ybot-top is 200, in 8 divisions, is 25 per division.
        X-axis has 8 division. A return of 600 points represents 50 pts/div.
        So a 600 pt data has a time step of TimeDiv/50.
        A(V) = (240-RawByte)*(Vdiv/25) - ( VertOffset + VoltDiv*4.6 )
            Where (Vdiv/25) is yincrement.
        T(s) = (PtNum-1)*(TimeDiv/50) - ( TimeDiv*6 - TimeOffset )
            Where (TimeDiv/50) is xincrement.
        BUG: This sometimes fails reading LONG memory. Currently this appear to be related to the digital filter. Ensure it is off.
        """
        # BUG: Make false 1st call to avoid above bug.
        if not hasattr(self, "_waveform_1st_call"):
            self._waveform_1st_call = []
        if chan not in self._waveform_1st_call:
            self._waveform_1st_call.append(chan)
            # Ensure digital filter off. Might not change in STOP mode, so just abort.
            self.channel_filter(chan, "OFF")
            if self.channel_filter(chan, query = True) != "OFF":
                raise visacomsbase.VisaComsBaseError("Please disable digital filter.")
            # Tried using self.instrument.query_binary_values, but write and read_raw seem to clear bug.
            #self.instrument.write(":WAV:DATA? " + chan)
            #nums = self.instrument.read_raw()
            nums = self.instrument.query_binary_values(":WAV:DATA? " + chan, datatype = 'B', container = numpy.array)
        nums = self.instrument.query_binary_values(":WAV:DATA? " + chan, datatype = 'B', container = numpy.array)
        return nums
    
    def waveform_preamble(self, chan):
        """Build a synthetic preamble data structure to match DS1000Z scope. Only key parameters added. 
        - chan - CHAN1|CHAN2
        Wherein,
        <format>: 0 (BYTE), 1 (WORD) or 2 (ASC).
        <type>: 0 (NORMal), 1 (MAXimum) or 2 (RAW).
        <points>: an integer between 1 and 12000000.
        <count>: the number of averages in the average sample mode and 1 in other modes.
        <xincrement>: the time difference between two neighboring points in the X direction.
            NORM: xincrement = TimeScale/100. RAW: xincrement = 1/SampleRate
        <xorigin>: the time from the trigger point to the "Reference Time" in the X direction.
        <xreference>: the reference time of the data point in the X direction.
        <yincrement>: the waveform increment in the Y direction.
        <yorigin>: the vertical offset relative to the "Vertical Reference Position" in the Y direction.
        <yreference>: the vertical reference position in the Y direction.
        Additional items,
        samplerate: the sample rate, perhaps this is generally 1.0/xincrement.
        xscale: the time base per division.
        yscale: the channel scale per division.
        """
        preamble = dataholder.DataHolder("DSwavedata")
        # Additions items.
        preamble.samplerate = self.acquire_sampling_rate()
        preamble.xscale = self.timebase_scale(query = True)
        preamble.yscale = self.channel_scale(chan, query = True)
        # Items that match DS1000Z scope.
        preamble.format = 0
        preamble.type = self.waveform_mode(query = True)
        preamble.points = self.channel_memory_depth(chan)
        preamble.count = 1
        preamble.xincrement = 1.0 / preamble.samplerate
        preamble.xorigin = 0.0
        preamble.xreference = self.timebase_offset(query = True)
        preamble.yincrement = preamble.yscale / 25.0 
        preamble.yorigin = 0.0
        preamble.yreference = self.channel_offset(chan, query = True)
        return preamble
