#!/usr/bin/env python
"""Digital Oscilloscope interface - USB.

Rigol DS1000Z series digital oscilloscopes.
Validated with instrument.

Note this is not a full implementation. Just the necessary and basic's functions.

Author: Ian Watson (igw)
Date: 6 Jan 2015
"""

import numpy

import visacomsbase
import DS_Common
import dataholder


################################################################################
# Channels by name.
CHANOFF = "OFF"
CHAN1 = "CHAN1"
CHAN2 = "CHAN2"
CHAN3 = "CHAN3"
CHAN4 = "CHAN4"
CHANMATH = "MATH"
# Channel names. By index, 0 is OFF, Chan 1-4. 5 is MATH.
CHANNAMES = ( CHANOFF, CHAN1, CHAN2, CHAN3, CHAN4, CHANMATH )

class DS_DS1000Z(DS_Common.DS_Common):
    """Interface to digital oscilloscope."""
    
    # Default resource name.
    resource_name = "USBInstDS1000Z"
    # Dict of arguments for visa interface.
    # Takes long time to read 1Mpt long memory, so increase timeout.
    args_visa = { "timeout" : 15000, "chunk_size" : 2500000, "read_termination" : "\n" }
    # Dict of arguments for visacoms module.
    # This instrument has a IVI USB interface. It seems to hold the bus while a command is executed, so no delays appear to be required.
    args_visacoms = {
        "what_am_i" : "I might be a Rigol DS1000Z series digital oscilloscope.",
        "prompt" : "DS_DS1000Z",
        "write_delay" : 0.1
        #"read_term_chars_strip" : '\n\r'
    }
    # This instrument say it respond to OPC?
    
    # Inherited __init__
    
    # Channels.
    CHAN1 = CHAN1
    CHAN2 = CHAN2
    CHAN3 = CHAN3
    CHAN4 = CHAN4
    # All channels.
    CHANS = ( CHAN1, CHAN2, CHAN3, CHAN4 )
    # Device parameters.
    DIVISIONSY = 8
    DIVISIONSX = 12
    # All available y-scales.
    YDIVSCALES = [0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5, 10, 20, 50, 100]
    # Range of scales (inclusive) for each probe attenuation. Index to YDIVSCALES.
    YDIVSCALEBYPROBE = {"1" : (0, 12), "5" : (2, 14), "10" : (3, 15)}
    #
    time_div_max = 0.01
    
    ############################################################################
    # Acquisition commands.
    def system_run(self):
        """Starts continuous acquisitions until STOP or SING."""
        self.trigger_sweep_mode("AUTO")
        self.write(":RUN")
    
    def system_stop(self):
        """Stops the running acquisition. MAY TRUNCATE DATA CAPTURE. Use system_single.
        OK this is data safe on Rigol, but not on Hameg.
        """
        self.write(":STOP")
    
    def system_single(self):
        """Starts (it triggers) a defined number of acquisition cycles.
        This does trigger a single sweep to start (restart if already running).
        The scope appear to STOP at the end of the sweep. BUT.
        You should wait until the sweep has finished, with say 2s margin, then issue system_stop().
        Then data can be pulled from the scope reliably.
        """
        self.write(":SING")
        #self.trigger_sweep_mode("SING")
        #self.trigger_force()
        # HACK. Extra delay to ensure SING has finished. Or else preamble is junk.
        #self.sleep_em(0.1)
        #self.system_stop()
        #self.sleep_em(0.1)
    
    ########
    # :ACQuire
    def acquire_memory_depth(self, mode = None, query = False, debug = "AUTO"):
        """Set/query acquire memory depth = AUTO|number
        In points.
        1ch = AUTO|12000|120000|1200000|12000000|24000000
        2ch = 1ch / 2
        4ch = 1ch / 4
        memDepth = SampRate x WaveformLength, WaveformLength = 12xTimeBaseScale
        """
        mess = self.arg1_with_query(":ACQ:MDEP", None, mode, query, debug = debug)
        return mess
    
    def acquire_type(self, mode = None, query = False, debug = "NORM"):
        """Set/query acquire type = NORMal|AVERages|PEAK|HRESolution"""
        mess = self.arg1_with_query(":ACQ:TYPE", ["NORM", "AVER", "PEAK", "HRES"], mode, query, debug = debug)
        return mess
    
    def acquire_aver_count(self, mode = None, query = False, debug = "16"):
        """Set/query average count. 2, 4, 8, to 1024, i.e., 2^N, N = 1-10"""
        mess = self.arg1_with_query("ACQ:AVER", None, mode, query, debug = debug)
        if query:
            return int(mess)
    
    def acquire_aver_wait_till_done(self):
        """Query, waits for average count to complete.
        FIXME: hack to get this working.
        """
        if self.acquire_type(query = True) == "AVER":
            ts = self.timebase_scale(query = True)
            ave_num = self.acquire_aver_count(query = True)
            self.sleep_em( ave_num * (ts + 0.001) * self.DIVISIONSX )
        
        return 1
    
    def acquire_sampling_rate(self, debug = 100e6):
        """Query acquire sampling rate query.
        memDepth = SampRate x WaveformLength, WaveformLength = 12xTimeBaseScale
        """
        mess = self.get_query(":ACQ:SRAT", debug = debug)
        return float(mess)
    
    ########
    # :CHANnel
    def channel_bandwidth_limit(self, chan, mode = None, query = False, debug = "OFF"):
        """Set/query channel bandwidth limit = 20M|OFF
        - chan - CHAN1|CHAN2|CHAN3|CHAN4
        """
        mess = self.arg1_with_query(":" + chan + ":BWL", ["20M", "OFF"], mode, query, debug = debug)
        return mess
    
    def channel_coupling(self, chan, mode = None, query = False, debug = "DC"):
        """Set/query channel coupling = DC|AC|GND
        - chan - CHAN1|CHAN2|CHAN3|CHAN4
        """
        mess = self.arg1_with_query(":" + chan + ":COUP", ["DC", "AC", "GND"], mode, query, debug = debug)
        if not query and mode == "AC":
            self.sleep_em(1.2)
        return mess
    
    def channel_display(self, chan, mode = None, query = False, debug = "OFF"):
        """Set/query channel display = ON|OFF
        - chan - CHAN1|CHAN2|CHAN3|CHAN4
        """
        mess = self.arg1_with_query(":" + chan + ":DISP", ["ON", "OFF"], mode, query, debug = debug)
        return mess
    
    def channel_offset(self, chan, mode = None, query = False, debug = "0"):
        """Set/query channel offset = number
        - chan - CHAN1|CHAN2|CHAN3|CHAN4
        vertical scale>500mV/div: -100V to +100V
        vertical scale<500mV/div: -2V to +2V
        """
        mess = self.arg1_with_query(":" + chan + ":OFFS", None, mode, query, debug = debug)
        if query:
            return float(mess)
    
    def channel_offset_position(self, chan, mode = None, query = False, debug = "0"):
        """Dummy function. Set/query channel offset position (in divisions) = -5 <= divisions <= 5
        - chan - CHAN1|CHAN2|CHAN3|CHAN4
        """
        scale = self.channel_scale(chan, query = True)
        if query:
            mess = self.channel_offset(chan, mode, query, debug = debug) / scale
            return mess
        mode = mode * scale
        self.channel_offset(chan, mode, query)
    
    def channel_probe_type(self, chan):
        """Dummy function. Query channel probe type = NONE | ACTive | PASSive
        - chan - CHAN1|CHAN2|CHAN3|CHAN4
        """
        return "NONE"
    
    def channel_probe(self, chan, mode = None, query = False, debug = "10"):
        """Set/query channel probe (attenuation factor) = 0.01, 0.02, 0.05, 0.1 to 1000
        - chan - CHAN1|CHAN2|CHAN3|CHAN4
        """
        mess = self.arg1_with_query(":" + chan + ":PROB", None, str(mode), query, debug = debug)
        if query:
            # Returns an string exponent format, ex., '1.000000e+01'. Convert to tidy sting.
            return str( int( float(mess) ) )
    
    def channel_range(self, chan, mode = None, query = False, debug = "8"):
        """Set/query channel range (8xscale per division) = number
        x1 is 2mV to 10V. x10 is 20mV to 100V.
        - chan - CHAN1|CHAN2|CHAN3|CHAN4
        """
        mess = self.arg1_with_query(":" + chan + ":RANG", None, mode, query, debug = debug)
        if query:
            return float(mess)
    
    def channel_scale(self, chan, mode = None, query = False, debug = "1"):
        """Set/query channel scale (per division) = number
        x1 is 1mV to 10V. x10 is 10mV to 100V.
        Can be any value if vernier on.
        - chan - CHAN1|CHAN2|CHAN3|CHAN4
        """
        mess = self.arg1_with_query(":" + chan + ":SCAL", None, mode, query, debug = debug)
        if query:
            return float(mess)
    
    def channel_venier(self, chan, mode = None, query = False, debug = "OFF"):
        """Set/query channel scale vernier = ON|OFF
        - chan - CHAN1|CHAN2|CHAN3|CHAN4
        """
        mess = self.arg1_with_query(":" + chan + ":VERN", ["ON", "OFF"], mode, query, debug = debug)
        return mess
    
    ########
    # :DISPlay
    def display_bitmap(self, mode = None, query = False):
        """Query a bitmap of the display grid.
        FIXME: TODO: this will probably need debug on the instrument
        """
        self.write(":DISP:DATA?")
        mess = self.read()
        assert mess[0] == "#"
        nchars = mess[1]
        len = int( mess[2:2+nchars] )
        # Change visa to binary
        bitmap = self.instrument.read(len)
        # Remove 54 bytes of bitmap header.
        bitmap = bitmap[54:]
        return mess
    
    def display_grid(self, mode = None, query = False, debug = "0"):
        """Set/query display grid = FULL|HALF|NONE"""
        mess = self.arg1_with_query(":DISP:GRID", ["FULL", "HALF", "NONE"], mode, query, debug = debug)
        return mess
    
    ########
    # :MATH
    def math_display(self, mode = None, query = False, debug = "OFF"):
        """Set/query math display = ON|OFF"""
        mess = self.arg1_with_query(":MATH:DISP", ["ON", "OFF"], mode, query, debug = debug)
        return mess
    
    def math_operate(self, mode = None, query = False, debug = "ADD"):
        """Set/query math operate =  ADD|SUBTract|MULTiply|DIVision|AND|OR|XOR|NOT|FFT|INTG|DIFF|SQRT|LOG|LN|EXP|ABS
        """
        mess = self.arg1_with_query(":MATH:OPER", ["ADD", "SUBT", "MULT", "DIV", "AND", "OR", "XOR", "NOT", "FFT", "INTG", "DIFF", "SQRT", "LOG", "LN", "EXP", "ABS"], mode, query, debug = debug)
        return mess
    
    #def fft_display(self, mode = None, query = False):
    #    """Set/query FFT display = ON|OFF"""
    #    mess = self.arg1_with_query(":FFT:DISP", ["ON", "OFF"], mode, query)
    #    return mess
    
    ########
    # :MEAS
    def measure_counter_source(self, mode = None, query = False, debug = "CHAN1"):
        """Set/query measurement counter source = CHANnel1|CHANnel2|CHANnel3|CHANnel4|OFF"""
        mess = self.arg1_with_query(":MEAS:COUN:SOUR", ["CHAN1", "CHAN2", "CHAN3", "CHAN4", "OFF"], mode, query, debug = debug)
        return mess
    
    def measure_counter_value(self, debug = 1e6):
        """Query measurement counter value"""
        mess = self.get_query(":MEAS:COUN:VAL", debug = debug)
        return float(mess)
    
    def measure_source(self, mode = None, query = False, debug = "CHAN1"):
        """Set/query measurement source = CHANnel1|CHANnel2|CHANnel3|CHANnel4
        Return CHAN1, ...
        """
        mess = self.arg1_with_query(":MEAS:SOUR", ["CHAN1", "CHAN2", "CHAN3", "CHAN4"], mode, query, debug = debug)
        return mess
    
    def measure_item(self, mode, chan = None, chanref = None, munit = 6, or_repeat = 8, query = False, debug = 1):
        """Set/query measurement item of channel(s),
        VMAX | VMIN | VPP | VTOP | VBASe | VAMP | VAVG | VRMS | OVERshoot | PREShoot | MARea | MPAReaPERiod | FREQuency | RTIMe | FTIMe | PWIDth | NWIDth | PDUTy | NDUTy | RDELay | FDELay | RPHase | FPHase
        
        Sets measure item of channel or channels.
        If query, sets and returns result.
        Returns '9.9e37' if over level.
        
        FIXME: TODO: Check if must be set before query.
        - chan - str of channel or list of str's of channels. i.e. 'CHAN1' or ['CHAN3', 'CHAN4']
        - chan - str of source, usually channel, say CHAN1.
        - chanref - 2nd channel reference.
        - munit - Measurement unit if appropriate. Not used, for compat.
        - or_repeat - Overrange/overload repeat measurement count.
        """
        if self.emulation:
            return debug
        out = ":MEAS:ITEM" + ( query is False and " " or "? " ) + mode
        if chan is not None:
            if type(chan) is not list:
                chan = [chan]
            for ch in chan:
                out += "," + ch
        #self.write(out)
        if not query:
            self.write(out)
            return
        
        # Do measurement.
        mess = self._measure_item_cmd(out)
        
        if mess >= 1e37 and or_repeat > 0:
            print(">> Measure over range. Retry %d times" % ( or_repeat ), end = '')
            for count_n in range(or_repeat):
                mess = self._measure_item_cmd(out)
                print(", %g" % (mess), end = '')
                if mess < 1e37:
                    break
            print(".")
            if mess >= 1e37:
                print(">> Retry unsuccessful.")
        
        return mess
    
    def _measure_item_cmd(self, out):
        """The measure item command."""
        self.write(out)
        # I think a little wait can only help here.
        #self.sleep_em(0.2)
        mess = self.read()
        
        return float(mess)
    
    ########
    # :TIMebase
    def timebase_mode(self, mode = None, query = False, debug = "MAIN"):
        """Set/query timebase mode = MAIN|XY|ROLL"""
        mess = self.arg1_with_query(":TIM:MODE", ["MAIN", "XY", "ROLL"], mode, query, debug = debug)
        return mess
    
    def timebase_offset(self, mode = None, query = False, debug = "0"):
        """Set/query timebase offset = number"""
        mess = self.arg1_with_query(":TIM:OFFS", None, mode, query, debug = debug)
        if query:
            return float(mess)
    
    def timebase_scale(self, mode = None, query = False, debug = "1e-6"):
        """Set/query timebase scale (per division) = number of 5ns - 50s
        Time will be rounded to nearest 1,2,5 notation by scope.
        """
        if mode:
            mode = "%0.9f" % mode
        mess = self.arg1_with_query(":TIM:SCAL", None, mode, query, debug = debug)
        if query:
            return float(mess)
    
    ########
    # :TRIGger
    def trigger_type(self, mode = None, query = False, debug = "EDGE"):
        """Set/query trigger mode = EDGE + others
        ONLY implementing EDGE.
        """
        mess = self.arg1_with_query(":TRIG:MODE", ["EDGE"], mode, query, debug = debug)
        if not query:
            self._trig_type = mode
        return mess
    
    SCPI_MAP_TRIGGER_COUPLING = visacomsbase.bidirectional_name_map( {"LF" : "LFR", "HF" : "HFR"} )
    def trigger_edge_coupling(self, mode = None, query = False, debug = "DC"):
        """Set/query trigger coupling = DC|AC|LF|HF - actual DC|AC|LFReject|HFReject"""
        mode = self.SCPI_MAP_TRIGGER_COUPLING.get(mode, mode)
        mess = self.arg1_with_query(":TRIG:COUP", ["DC", "AC", "LFR", "HFR"], mode, query, debug = debug)
        return self.SCPI_MAP_TRIGGER_COUPLING.get(mess, mess)
    
    def trigger_edge_coupling_noise_reject(self, mode = None, query = False, debug = "OFF"):
        """Set/query trigger edge noise reject = ON|OFF"""
        mess = self.arg1_with_query("TRIG:NREJ", ["ON", "OFF"], mode, query, debug = debug)
        return mess
    
    def trigger_status(self, debug = "AUTO"):
        """Query trigger status. Return TD, WAIT, RUN, AUTO, STOP."""
        mess = self.get_query(":TRIG:STAT", debug = debug)
        return mess
    
    def trigger_sweep_mode(self, mode = None, query = False, debug = "AUTO"):
        """Set/query trigger sweep mode = AUTO|NORMal|SINGle
        Use system_single() for a single sweep.
        Valid data will be available after single sweep is triggered or force trigger given.
        To resume continuous capture, change back to AUTO|NORM, or system_run().
        """
        mess = self.arg1_with_query(":TRIG:SWE", ["AUTO", "NORM", "SING"], mode, query, debug = debug)
        return mess
    
    # For EDGE
    def trigger_alltype_level(self, mode = None, query = False, debug = "0"):
        """Set/query trigger oftype level = voltage level number"""
        mess = self.arg1_with_query(":TRIG:" + self._trig_type + ":LEV", None, mode, query, debug = debug)
        if query:
            return float(mess)
    
    def trigger_alltype_source(self, mode = None, query = False, debug = "CHAN1"):
        """Set/query trigger oftype source = CHANnel1|CHANnel2|CHANnel3|CHANnel4|AC"""
        mess = self.arg1_with_query(":TRIG:" + self._trig_type + ":SOUR", ["CHAN1", "CHAN2", "CHAN3", "CHAN4", "AC"], mode, query, debug = debug)
        return mess
    
    def trigger_force(self):
        """Trigger force trigger. For NORM|SING mode."""
        self.write(":TFORC")
    
    ########
    # :WAVeform
    def waveform_source(self, mode = None, query = False, debug = "CHAN1"):
        """Set/query waveform data source = CHANnel1|CHANnel2|CHANnel3|CHANnel4|MATH
        If MATH, only NORMal data available.
        """
        mess = self.arg1_with_query(":WAV:SOUR", ["CHAN1", "CHAN2", "CHAN3", "CHAN4", "MATH"], mode, query, debug = debug)
        return mess
    
    def waveform_mode(self, mode = None, query = False, debug = "MAX"):
        """Set/query waveform reading mode = NORMal|MAXimum|RAW
        NORM is data displayed on screen, 1200 points, 100 points per 12 divisions.
        MAX screen when running, memory when stopped.
        RAW is memory, but must be in STOP state to use this.
        """
        mess = self.arg1_with_query(":WAV:MODE", ["NORM", "MAX", "RAW"], mode, query, debug = debug)
        return mess
    
    def waveform_format(self, mode = None, query = False, debug = "ASC"):
        """Set/query waveform return format = BYTE|WORD|ASCii
        This function is not generally called. See waveform_get_data().
        BYTE: one point per byte (8 bits).
        WORD: one point per 2 bytes, lower byte valid, higher byte is 0.
        ASC: points in character format, comma separated.
        """
        mess = self.arg1_with_query(":WAV:FORM", ["BYTE", "WORD", "ASC"], mode, query, debug = debug)
        return mess
    
    def waveform_get_data(self, chan = None, points = None, mode = "MAX"):
        """Get waveform data from source, chan.
        The RUN/STOP status will affect the mode data length returned. Set RUN/STOP before calling.
        Returns a dataholder.DataHolder structure containing all the waveform_preamble items and data items data_x & data_y.
        The data is converted to real units, usually time and voltage.
        - chan - CHANnel1|CHANnel2|CHANnel3|CHANnel4|MATH
        - points - Number of points to return, or maximum available.
        - mode - NORMal|MAXimum|RAW
        
        http://www.eevblog.com/forum/testgear/rigol-ds1000z200040006000-waveform-capture-utility
        Rigol DS1074z Bugs:
        When using :WAV:DATA? to request samples there is an upper limit of exactly 1179644 data points that can be downloaded at any one time requesting more points using :WAV:START and :WAV:STOP will result in a no data being returned at all.
        Note this is when downloading the data as bytes ie :WAV:FORM BYTE, if using :WAV:FORM ASC (ascii) the limit is reduced to around 60Kpoints.
        However it is possible to download the entire waveform memory in blocks and this is the technique I have used to download more than 1.1Mpoints.
        """
        # Set data source, mode and format.
        if chan is not None:
            self.waveform_source(chan)
        
        #mode_query = self.waveform_mode(query = True)
        #print "waveform_mode =", mode_query
        self.waveform_mode(mode)
        
        # The read mode. Only BYTE is worth while using.
        #wformat = "ASC"
        wformat = "BYTE"
        #wformat = "WORD"
        self.waveform_format(wformat)
        
        # Change data start/stop points to memory depth.
        self.write(":WAV:START 1")
        mem_depth = self.acquire_memory_depth(query = True)
        self.write(":WAV:STOP " + mem_depth)
        
        # Get preamble data structure. Modify to suit data below.
        wavedata = self.waveform_preamble(chan)
        
        # Get mem depth and loop over acquiring it in chunks. Limit
        mem_depth = int( self.acquire_memory_depth(query = True) )
        # If screen mode NORM, then only 1200 point will be available.
        mem_depth = min(wavedata.points, mem_depth)
        if points is not None:
            mem_depth = min(mem_depth, points)
        mem_step = min( {"BYTE":250000, "WORD":125000, "ASC":15625}[wformat], mem_depth )
        mem_got = 0
        wavedata.data_y = []
        while mem_got < mem_depth:
            # Set memory window to acquire.
            self.write( ":WAV:START " + str(1 + mem_got) )
            mem_stop = min(mem_got + mem_step, mem_depth)
            self.write( ":WAV:STOP " + str(mem_stop) )
            
            # Request data and read. Wait a prescribed time before read to allow instrument to assembly data. TODO: Validate if delay really required. Manual recommends delay; 0.2s bytes, 0.3s words, 2s ascii, but cannot spot this now.
            self.write(":WAV:DATA?")
            self.sleep_em( {"BYTE":0, "WORD":0, "ASC":1.5}[wformat] )
            if wformat == "ASC":
                mess = self.read()
                # First 11 are header, e.g., #900000dddd.
                mess = mess[11:]
                nums = mess.split(",")
                nums = numpy.array(nums, dtype = numpy.float)
            else:
                # Read BYTE & WORD. Seems the TMS header handles this correctly with normal read.
                mess = self.instrument.read_raw()
                # First 11 are header & length as dddd, e.g., #900000dddd. Includes '\n' at end due to read_raw().
                mess = mess[11:-1]
                #print "ord(0-N)",
                #for n in range(4):
                #    print(ord(mess[n]), end=' ')
                #print
                if wformat == "BYTE":
                    # Convert str byte chars to decimal list (now as array). Add 0.0 to convert to float.
                    nums = numpy.fromstring(mess, dtype = ">u1") + 0.0
                else:
                    # Convert str chars, 2 per unsigned word, to decimal list (now as array).
                    nums = numpy.fromstring(mess, dtype = "<u2") + 0.0
                # Change y-axis data based on wavedata information.
                nums = ( nums - wavedata.yreference - wavedata.yorigin ) * wavedata.yincrement
            
            # Append y-axis data.
            wavedata.data_y = numpy.append(wavedata.data_y, nums)
            # Current length for loop over data length.
            mem_got = len(wavedata.data_y)
        
        # Ensure correct length.
        wavedata.points = len(wavedata.data_y)
        # Create x-axis data.
        wavedata.data_x = numpy.arange(0, wavedata.points) * wavedata.xincrement
        #wavedata.data_x = numpy.linspace(0, wavedata.points*wavedata.xincrement, wavedata.points, endpoint = False)
        
        # Ensure ASC as this may affects other reads.
        self.waveform_format("ASC")
        
        return wavedata
    
    def waveform_preamble(self, chan):
        """Query waveform preamble data structure.
        - chan - CHAN1|CHAN2|CHAN3|CHAN4
        Wherein,
        <format>: 0 (BYTE), 1 (WORD) or 2 (ASC).
        <type>: 0 (NORMal), 1 (MAXimum) or 2 (RAW).
        <points>: an integer between 1 and 12000000.
        <count>: the number of averages in the average sample mode and 1 in other modes.
        <xincrement>: the time difference between two neighboring points in the X direction.
            NORM: xincrement = TimeScale/100. RAW: xincrement = 1/SampleRate
        <xorigin>: the time from the trigger point to the "Reference Time" in the X direction.
        <xreference>: the reference time of the data point in the X direction.
        <yincrement>: the waveform increment in the Y direction.
        <yorigin>: the vertical offset relative to the "Vertical Reference Position" in the Y direction.
        <yreference>: the vertical reference position in the Y direction.
        Additional items,
        samplerate: the sample rate, perhaps this is generally 1.0/xincrement.
        xscale: the time base per division.
        yscale: the channel scale per division.
        """
        # <format>,<type>,<points>,<count>,<xincrement>,<xorigin>,<xreference>,<yincrement>,<yorigin>,<yreference>
        mess = self.get_query(":WAV:PRE")
        mess = mess.split(",")
        preamble = dataholder.DataHolder("DSwavedata")
        # TODO: Change to two list iterators over required indexes.
        preamble.format = int( mess[0] )
        preamble.type = int( mess[1] )
        preamble.points = int( mess[2] )
        preamble.count = int( mess[3] )
        preamble.xincrement = float( mess[4] )
        preamble.xorigin = float( mess[5] )
        preamble.xreference = float( mess[6] )
        preamble.yincrement = float( mess[7] )
        preamble.yorigin = int( mess[8] )
        preamble.yreference =  int( mess[9] )
        # Additions items.
        preamble.samplerate = self.acquire_sampling_rate()
        preamble.xscale = self.timebase_scale(query = True)
        preamble.yscale = self.channel_scale(chan, query = True)
        return preamble
