#!/usr/bin/env python
"""PSU interface - ethernet, GPIB, RS232.

TTi PL 068, 155, 303, 601, 303QMD, 303QMT -P PSU.
Validated with instrument.

TODO: Perhaps rename this to be PLSeriesP.

Author: Ian Watson (igw)
Date: 21 Jul 2015
"""

import visacomsbase


################################################################################
# Output channels.
CH1 = 1
CH2 = 2
CH3 = 3
CH_MASTER = CH1
CH_SLAVE = CH2
CH_THIRD = CH3


class PSU_PL303QMDP(visacomsbase.VisaComsBase):
    """Interface to PSU."""
    
    # Default resource name.
    resource_name = "ASRL16"
    #resource_name = "COM16"
    # Dict of arguments for visa interface. Transmit to instrument terminator, '\n' = 0x10 [Line Feed, LF]. Replies with '\r\n' = 0x0D 0x10 [CR, LF].
    args_visa = { "write_termination" : "\n", "read_termination" : "\r\n" }
    # Dict of arguments for visacoms module.
    args_visacoms = {
        "what_am_i" : "I might be a TTi PL303QMD-P PSU.",
        "prompt" : "PSU_PL303QMDP",
        "write_delay" : 0.1
        #"read_term_chars_strip" : '\r\n'
    }
    # This instrument does respond to OPC?, but is of little use.
    
    # Inherited __init__
    
    # Output channels.
    CH1 = CH1
    CH2 = CH2
    CH3 = CH3
    CH_MASTER = CH1
    CH_SLAVE = CH2
    CH_THIRD = CH3
    
    ############################################################################
    # Voltage
    def conf_volt(self, ch, mode = None, query = False, debug = "1"):
        """Set/query voltage on channel = voltage number
        ch - 1|2
        """
        mess = self.arg1_with_query("V" + str(ch), None, mode, query, debug = "V " + debug)
        if query:
            return float( mess.split()[-1] )
    
    def conf_volt_over(self, ch, mode = None, query = False, debug = "0.1"):
        """Set/query over voltage protection on channel = voltage number
        ch - 1|2
        """
        mess = self.arg1_with_query("OVP" + str(ch), None, mode, query, debug = "I " + debug)
        if query:
            return float( mess.split()[-1] )
    
    # DELTAV
    # INCV
    # INCI
    # RATIO
    
    ########
    # Current
    def conf_curr(self, ch, mode = None, query = False, debug = "0.1"):
        """Set/query current limit on channel = current number
        ch - 1|2
        """
        mess = self.arg1_with_query("I" + str(ch), None, mode, query, debug = "I " + debug)
        if query:
            return float( mess.split()[-1] )
    
    def conf_curr_over(self, ch, mode = None, query = False, debug = "2"):
        """Set/query over current protection on channel = current number
        Above this the channel switches off its output.
        ch - 1|2
        """
        mess = self.arg1_with_query("OCP" + str(ch), None, mode, query, debug = "I " + debug)
        if query:
            return float( mess.split()[-1] )
    
    def conf_curr_range(self, ch, mode = None, query = False, debug = "1"):
        """Set/query current range on channel = 1 | 2, for, 500mA | 3A.
        ch - 1|2
        Return int 1 | 2 (as above)
        """
        mess = self.arg1_with_query("IRANGE" + str(ch), ["1", "2"], mode, query, debug = debug)
        if query:
            return int(mess)
    
    # DELTAI
    # INCI
    # DECI
    
    ########
    # Channel
    def channel_output(self, ch, mode = None, query = False, debug = "0"):
        """Set/query channel output on/off = OFF | ON | 0 | 1
        ch - 1|2
        Return 0 | 1
        """
        mess = self.arg1_with_query("OP" + str(ch), None, self.map2num_bool(mode), query, debug = debug)
        if query:
            return int(mess)
    
    def channel_output_all(self, mode):
        """Set all channels output on/off = OFF | ON | 0 | 1
        """
        self.arg1_with_query("OPALL", None, self.map2num_bool(mode) )
    
    def channel_volt(self, ch, debug = "1"):
        """Query channel voltage readback from output
        ch - 1|2
        Instrument return '<number_value>V'.
        Not a meaningful results if channel output not active.
        """
        mess = self.get_query("V" + str(ch) + "O", debug = debug + "V")
        return float( mess[:-1] )
    
    def channel_curr(self, ch, debug = "0.1"):
        """Query channel current readback from output
        ch - 1|2
        Instrument return '<number_value>A'.
        Not a meaningful results if channel output not active.
        """
        mess = self.get_query("I" + str(ch) + "O", debug = debug + "A")
        return float( mess[:-1] )
