#!/usr/bin/env python
"""Trivial data holder for various instruments.

Author: Ian Watson (igw)
Date: 6 Jul 2012
"""

#import time
import copy
#import math
import numpy
import matplotlib.pyplot as plt


################################################################################
class DataHolder(object):
    """A trivial base class for holding instruments data items.
    User should attach attributes as required.
    """
    
    def __init__(self, dataType = "DataHolder"):
        """- dataType - String indicating data type for Human.
        Usual properties,
        .points
        .data_x
        .data_y
        """
        self.dataType = dataType
    
    ############################################################################
    # Utility functions.
    def copy(self):
        """Make a copy of the current object."""
        return copy.copy(self)
    
    def check_data_len(self):
        """Check data_x/y len matches points."""
        assert len( self.data_x ) == self.points, "Length data_x " + str( len(self.data_x) ) + " != " + str(self.points)
        assert len( self.data_y ) == self.points, "Length data_y " + str( len(self.data_y) ) + " != " + str(self.points)
    
    def waveform_clip_len(self, points):
        """Reduce length of DSwavedata to required length."""
        self.points = min(self.points, points)
        self.data_x = self.data_x[0:self.points]
        self.data_y = self.data_y[0:self.points]
    
    def waveform_clip2prevpow10(self, points = None):
        """Reduce length of DSwavedata to closest power of 10.
        This provide ready for nice FFT.
        Ex. Data with say 1024 points will be reduced to 1000.
        """
        if points is None:
            L = self.points
        else:
            L = min(self.points, points)
        # int() truncates towards zero.
        L = 10 ** int( numpy.log10(L) )
        self.waveform_clip_len(L)
    
    ############################################################################
    # Plot functions.
    def waveform_plot(self, form = True):
        """Plot a waveform from 'DSwavedata' structure.
        - form - Allows plots to be disabled. "dB" for dB y-scale.
        """
        if form:
            if form == "dB":
                plt.plot(self.data_x, 20 * numpy.log10(self.data_y))
            else:
                plt.plot(self.data_x, self.data_y)
            plt.show()
    
    def waveform_plot_fft(self, form = True):
        """As _plot, but dots each point. Best for FFT.
        - form - Allows plots to be disabled. "dB" for dB y-scale.
        """
        if form:
            if form == "dB":
                plt.plot(self.data_x, 20 * numpy.log10(self.data_y), "-o")
            else:
                plt.plot(self.data_x, self.data_y, "-o")
            plt.show()
    
    ############################################################################
    # FFT utility.
    def waveform_fft(self):
        """Perform FFT on time-voltage waveform.
        Applies requested window before FFT.
        Remember amplitude will be correct with no window, and wrong with a window.
        The results are added to self,
        .data_x - Replaced with frequency.
        .data_y - Replaced with amplitude.
        .f_bin  - Frequency binning.
        Returns (f_samples, spec_amp, f_bin)
        TODO: Could add window arg for required hanning, hamming, etc.
        """
        self.check_data_len()
        L = self.points
        #if window is not None:
        #    self.data_y = self.data_y * numpy.hanning(L)
        
        # Do FFT and scale. Keep DC, index 0 the same. Double 1:halfLength.
        spec_amp = numpy.fft.fft( self.data_y ) / L
        self.data_y = numpy.absolute( spec_amp[0:L/2] ) * 2
        self.data_y[0] /= 2
        #self.data_y = numpy.fft.fftshift(self.data_y)
        # Frequency bin resolution and frequency axis.
        self.f_bin = 1.0 / self.xincrement / float(L)
        self.data_x = self.f_bin * numpy.arange(0, L/2)
        self.points = L/2
        # Append FFT to names of data.
        self.dataType += "FFT"
        return (self.data_x, self.data_y, self.f_bin)
    
    def waveform_fft_level_at_freq(self, freq):
        """Get level of FFT signal at given frequency. The frequency bin closest to the given is chosen."""
        f_samples = self.data_x
        #idxs = numpy.argwhere( f_samples > freq )
        #idx = idxs2[0, 0]
        idxs = numpy.where( f_samples > freq )
        idx = idxs[0][0]
        if idx > 0 and ( abs( f_samples[idx-1] - freq ) < abs( f_samples[idx] - freq ) ):
            idx -= 1
        level = self.data_y[idx]
        return level
    
    def waveform_fft_level_around_freq(self, freq):
        """Get level of FFT signal around given frequency. The frequency bin closest to the given is chosen.
        Level is determined from sum of power of adjacent terms.
        If freq exact this is -1 and +1 terms.
        Otherwise the -2 to +2 terms are looked at and the largest 3 terms summed.
        """
        f_samples = self.data_x
        idxs = numpy.where( f_samples >= freq )
        idx = idxs[0][0]
        if idx > 0 and ( abs( f_samples[idx-1] - freq ) < abs( f_samples[idx] - freq ) ):
            idx -= 1
        idx_start = max(0, idx-2)
        idx_stop = min(len(f_samples), idx+2)
        #23 idxs = list(range(idx_start, idx_stop + 1))
        idxs = range(idx_start, idx_stop + 1)
        freqs = f_samples[idxs]
        amps = self.data_y[idxs]
        amps.sort()
        level_all = numpy.sqrt( numpy.sum( numpy.power(amps, 2) ) )
        level = numpy.sqrt( numpy.sum( numpy.power(amps[-3:], 2) ) )
        return level
