#!/usr/bin/env python
"""Main Kate device control.

Provides KateIC,
    which is build on top of KateComs,
    which calls KateSPI to do the work.

KateSPI is a simple wrapper over robotusbiss.

Object and functions provide configuration and utilities for the KateIC device.

Author: Ian Watson (igw)
Date: 14 Jan 2014
"""

# Load modules.
import sys
#import os
import math
import time
#from collections import OrderedDict

from comsutils import chrs2ints, ints2chrs, ints2hexs, value2ints, dec2bin, dec2hex, neatBinStr
import regfields
import katespi
import katerig


# Add readline completion to python interpreter if interactive.
if sys.platform != "win32":
    import readline
    import rlcompleter
    # See readline man page for this.
    readline.parse_and_bind("set show-all-if-ambiguous on")
    readline.parse_and_bind("tab: complete")


################################################################################
class KateComsError(Exception):
    """KateComs Exception class."""
    pass


class KateComs(object):
    """KateIC communication module.
    This provide writeXY() & readXY() functions for a KateIC device.
    Two type of access are provided:-
    1 - Bit write/read to analog registers.
    2 - Bit write/read of banked digital registers.
    
    A flag to enable the actual communication is attached here.
    This prevent coms starting before registers have been built - the Gtk callbacks fire during the build process.
    
    Write always go to cache.
    When allowComs set, writes will go to bus. The flag cacheOnly will block this bus activity.
    
    Reads will be from bus unless allowComs 
    
    - phy - KateSPI providing the physical byte write & read interface.
    - allowComs - flag to enable I2C coms. Until set I2C coms is blocked.
    """
    # Class global attributes.

    def __init__(self, phy, allowComs = False, mappingFileName = None):
        """Initialisation."""
        # Coms Class object. Must be like KateSPI.write/read functions.
        self.phy = phy
        
        # Coms enable flag.
        self.allowComs = allowComs
        
        # Setup reg fields db.
        self._regdb = regfields.RegFields(mappingFileName)
    
    def __del__(self):
        """Close coms upon delete."""
        self.close()
    
    def close(self):
        """Close coms port."""
        if self.phy:
            self.phy.close()
    
    def enableComs(self):
        """Enables the coms. Following calls to write/read will work."""
        self.allowComs = True
    
    def disableComs(self):
        """Disable the coms. Following calls to write/read will NOT work."""
        self.allowComs = False
    
    def clearCaches(self):
        """Clear caches. Pass through to phy."""
        self.phy.clearCaches()
    
    ########
    # General register access. Handled Analog & Digital registers.
    def writeF(self, name, value, cacheWrite = False):
        """Write to register field.
        In emulation only cache is used.
        Bank access must ensure a read has been done before the write to ensure the cache is correct.
        - cacheWrite - Write to cache only.
        """
        if not self.allowComs:
            return None
        
        field = self._regdb.getField(name)
        
        # No point in stopping write to read only registers - it is allowed and also required for debug.
        #if field.rw == 'R':
        #    raise KateComsError("Attempt to write to read register. TODO Perhaps done prevent this.")
        
        # Check write value within range.
        if value < field.valueMin or value > field.valueMax:
            raise KateComsError("Write to '" +field.name  + "' outside field range, " + str(field.valueMin) + ":" + str(field.valueMax) + ".")
        # Read register ready for bit masking.
        cvs = self.phy.read(field.addr, (field.bitTotal + 7) // 8, True)
        cv = self.phy.int8s2value(cvs)
        # Handle valueOffset.
        value += field.valueOffset
        # Mask out field and logic or (add) with value bit aligned.
        nv = ( cv & field.bitMaskNot ) | ( ( value << field.bitOffset ) & field.bitMask )
        # Ensure correct number of bits.
        nv = nv & field.bitMaskTotal
        # Convert to bytes and send lower bytes first.
        values = self.phy.value2int8s(nv, field.bitTotal)
        self.phy.write(field.addr, values, cacheWrite = cacheWrite)
        return True
    
    def readReg(self, name, cacheRead = False):
        """Read the whole register containing given field. See readF() for details."""
        if not self.allowComs:
            return None
        
        field = self._regdb.getField(name)
        values = self.phy.read(field.addr, (field.bitTotal + 7) // 8, cacheRead)
        # If more than 1 byte read, reverse the order and convert to decimal.
        value = self.phy.int8s2value(values)
        return value
    
    def readF(self, name, cacheRead = False, tc = True):
        """Read from register field.
        In emulation only cache is used.
        - cacheRead - Read only from cache. NB emulation mode will always read from cache.
        - tc - returns value with two-complement adjustment if required.
        """
        if not self.allowComs:
            return None
        
        value = self.readReg(name, cacheRead)
        # Extract bit field.
        field = self._regdb.getField(name)
        value = ( value & field.bitMask ) >> field.bitOffset
        # Correct for valueOffset, i.e., two-complement numbers if required.
        if tc and value > field.valueMax:
            value -= field.valueOffset
        return value
    
    ########
    # Digital register access.
    def writeBank(self, name, value):
        """Write to digital register field and commit register."""
        self.writeF(name, value)
    
    def writeBankF(self, name, value):
        """Write to cached digital register field."""
        self.writeF(name, value, cacheWrite = True)
    
    def readBank(self, name, tc = True):
        """Read whole digital register containing given field.
        Pass any field within required register. Use readBankF() to access other fields from cache.
        """
        return self.readF(name, tc = tc)
    
    def readBankF(self, name, tc = True):
        """Read cached digital register field.
        tc - returns value with two-complement adjustment if required.
        """
        return self.readF(name, cacheRead = True, tc = tc)


def testKateComs():
    """Simple test KateComs."""
    # Create physical interface.
    phy = katespi.KateSPI(emulation = True)
    
    # Start coms. This loads register mapping file.
    coms = KateComs(phy, allowComs = True)
    
    name = "BGTRIM" # -8:7
    v = coms.readF(name)
    print("Read", name, " Got", v)
    for v in [7, -8]:
        print("Write", name, v)
        coms.writeF(name, v)
        v = coms.readF(name)
        print("Read", name, " Got", v)


################################################################################
class KateICError(Exception):
    """KateIC Exception class."""
    pass


################################################################################
# AUX ADC MUX's.
AUXADCMUX_RSSI = 0
AUXADCMUX_TSEN_TX = 1
AUXADCMUX_TSEN_BG = 2
AUXADCMUX_AUXADCXIN = 3
AUXADCMUX_POWSEN_V = 4
AUXADCMUX_POWSEN_I = 5
AUXADCMUX_VDDDIGx0P6 = 6
AUXADCMUX_VDDx0p4 = 7
AUXADCMUXS = ( "RSSI", "TSEN_TX", "TSEN_BG", "AUXADCXIN", "POWSEN_V", "POWSEN_I", "VDDDIGx0P6", "VDDPSx0p4" )
AUXADC_FS_VOLT = 1.75
AUXADC_FS_N_2C = 4096
AUXADC_FS_N_M = 8192

# RSSI detector
RSSIMUX_TXOUT = 0
RSSIMUX_TXMON = 1
RSSIMUX_RXIN = 2
RSSIMUX_GND = 3
RSSIMUXS = ( "TXOUT", "TXMON", "RXIN", "GND" )

# Analog regulators setting index.
REG_ANA_VOLTAGES = (2.70, 2.66, 2.62, 2.58,
                    2.86, 2.82, 2.78, 2.74,
                    3.02, 2.98, 2.94, 2.90,
                    3.18, 3.14, 3.10, 3.06
)
REG_ANA_VOLTAGES_DICT = dict( enumerate(REG_ANA_VOLTAGES) )

# Digital regulators setting index. 1.5 to 2.1 in 0.04 step, for -8 to +7
REG_DIG_VOLTAGES = [ round(1.82 + 0.04 * v, 2) for v in range(-8, 7+1) ]
REG_DIG_VOLTAGES_DICT = dict( enumerate(REG_DIG_VOLTAGES, -8) )

# Kate block gains.
# RX FEG. Index with 'RXFEG'.
GAIN_RXFEG = (3, 9, 15, 21)
# RX VGA gain by index. Gain index 0 to 31 & 32 to 63, providing -6dB to 25dB (twice) to 56dB of actual gain. NB. Gain has 1dB steps, but setting 31 & 32 provide the same gain, so range 0-63 provides 62 dB of range. 
GAIN_RXGAIN = list(range(-6, -6+32)) + list(range(-6+31, -6+31+32))


################################################################################
# Run if required.
if __name__ == "__main__":
    testKateComs()
