#!/usr/bin/env python
"""Wrapper of robotusbiss for Kate

Provides bytes write & read access via SPI.

Keeps a cache of each devices register values. This is used for all reads unless forced not to do so.
For each register address, cache store data as list of bytes. This handles one or multi byte length registers.

IMPORTANT INFORMATION.
HACK: The read only addresses are currently hard coded here.

Author: Ian Watson (igw)
Date: 12 Jan 2014
"""

# Load modules.
import time
from robotusbiss import *


################################################################################
# Read only registers. Hard coded here.
READ_ONLY_REGISTERS = [59, 60, 61, 62, 63]


class KateSPI(RobotUSBISS):
    
    def __init__(self, resource_name = None, mode = MODE_SPI_0, baud_rate = 100000, i2cAddr = 0x50, addrInc = 1, emulation = False):
        """Initialisation."""
        RobotUSBISS.__init__(self, resource_name, mode = mode, baud_rate = baud_rate, i2cAddr = i2cAddr, addrInc = addrInc, emulation = emulation)
        
        # Emulation register caches. This is a dictionary.
        self.clearCaches()
        
        # Time of last write access.
        self._time_last_write = None
        
        #if not emulation:
            # Bit banging.
            # TODO: If bit banging interface to be used, then the pins must be configured.
            # TODO: write code. 
    
    def clearCaches(self):
        """Clear the caches.
        Call after a device reset to ensure no memory of previous states.  
        """
        self._regCache = {}
    
    def get_time_last_write(self, i2cAddr = None):
        """
        - i2cAddr - Passed in case this is an I2C interface.
        """
        return self._time_last_write
    
    def clear_time_last_write(self, i2cAddr = None):
        """
        - i2cAddr - Passed in case this is an I2C interface.
        """
        self._time_last_write = None
    
    # Register write and read.
    # FIXME: We might need something to mark byte or byte sequence from a bank write. For now just assume only single register writes. Probably have a separate function for bank writes.
    def write(self, regAddr, regValues, cacheWrite = False):
        """Write byte or bytes to register of given device.
        Assumes access is to one registers at a time.
        - regAddr - register address, int.
        - regValues - int or list of int's, or hex string, with/without leading '0x'.
        - cacheWrite - Write to cache only.
        Return, list of integers.
        The slave unit must support auto regAddr increment for multiple write to be of use.
        """
        # Ensure regValues is a list for cache.
        regValues = value2ints(regValues, True)
        # Save in cache.
        self._regCache[regAddr] = regValues
        
        # Log time of write.
        self._time_last_write = time.time()
        
        # Exit if emulation as no phy is available. Exit if cacheWrite write.
        if self.emulation or cacheWrite:
            return True
        
        # Create packet. D7 W(1)/R(0) flag + 1st byte address. Then data bytes.
        values = [ 0x80 | ( regAddr & 0x3F ) ] + regValues
        # Write value(s).
        spi_rx = self.sendReceiveSPIbytes(values)
        #spi_rx = self.sendReceiveSPIbytesByBitBang(values)
        spi_rx = list( spi_rx[1:] )
        return spi_rx
    
    def read(self, regAddr, n = 1, cacheRead = False):
        """Read byte or bytes from register of given device.
        Assumes access is to one registers at a time.
        - regAddr - register address, int.
        - n - number of bytes to read.
        - cacheRead - Read only from cache. NB emulation mode will always read from cache.
        Return, single integer or list of integers.
        The slave unit must support auto regAddr increment for multiple write to be of use.
        """
        # Catch emulation or cache read.
        if self.emulation or cacheRead:
            # Get reg value or [0]'s if it does not exist.
            spi_rx = self._regCache.get(regAddr)
            if spi_rx is None:
                spi_rx = [0] * n
            #print "D-KateSPI> EMULATION read %d = %d" % ( regAddrCache, regValue )
        else:
            # Create packet. D7 W(1)/R(0) flag + 1st byte address. Then data bytes - send a dummy byte for each byte to read. Send 1 so visually the LSB marks the end of each byte.
            values = [ 0x00 | ( regAddr & 0x3F ) ] + [1] * n
            # Write and read value(s). Drop 1st byte read, which was the address.
            spi_rx = self.sendReceiveSPIbytes(values)
            #spi_rx = self.sendReceiveSPIbytesByBitBang(values)
            spi_rx = list( spi_rx[1:] )
        
        # Update cache with read data.
        self._regCache[regAddr] = spi_rx
        
        if n == 1:
            spi_rx = spi_rx[0]
        return spi_rx
    
    def value2int8s(self, value, nbits):
        """Convert integer value of length nbits, into list of int8's.
        Send lower bytes first.
        """
        # Send lower bytes first.
        values = value2ints( dec2hex(value, (nbits + 3) // 4), True )
        values.reverse()
        return values
    
    def int8s2value(self, nums):
        """Convert list of int8's to value.
        Lower byte is first.
        """
        if type(nums) == list:
            value = 0
            for num in reversed(nums):
                value = ( value << 8 ) + num
        else:
            value = nums
        return value
    
    ########
    # FALL BACK. If there is a problem with the SPI not interfacing with Kate correctly because of some subtle timing issue, a backup solution using bit banging can be switched in. Just switch the call function to the following.
    def sendReceiveSPIbytesByBitBang(self, values = []):
        """Send number of bytes to SPI, during which it will receives the same number of bytes.
        An error will occur if more than 63 are send at once.
        - values - list int8's to send. 1 to 63 off.
        Return, list int8's.
        TODO: Code the bit banging to do write and read.
        """
        self.logger.debug("Enter sendReceiveSPIbytes() -> %s", str(values) )
        if not values:
            return []
        #
        self._sendCmdBytes( [CMD_SPI_IO] + values )
        # Check return. Form: TX ACK/NACK, RX data (same length as TX).
        spi_rx = self._receiveCmdBytes( 1 + len(values) )
        if spi_rx[0] == CMDR_USBISS_NOTACK:
            raise RobotUSBISSError("sendReceiveSPIbytes write failed.")
        if spi_rx[0] != CMDR_USBISS_ACK:
            raise RobotUSBISSError("Unknown CMDR_USBISS_ACK status returned.")
        return spi_rx[1:]
