#!/usr/bin/env python
"""Simple data for Registers Fields. 
Provides access to register address and its field widths, offset, mask, etc., by field name.

Author: Ian Watson (igw)
Date: 14 Jan 2014
"""

# Load modules.
import sys


# Add readline completion to python interpreter if interactive.
if sys.platform != "win32":
    import readline
    import rlcompleter
    # See readline man page for this.
    readline.parse_and_bind("set show-all-if-ambiguous on")
    readline.parse_and_bind("tab: complete")


################################################################################
# File storing the detailed register field information.
MAPPING_FILE_NAME = "regMappingFile.txt"


class RegFieldError(Exception):
    """RegFields Exception class."""
    pass


class RegField(object):
    """Class holding attributes of each register field."""

    def __init__(self, name, rw, addr, bitTotal, bitWidth, bitOffset, valueMin, valueMax, valueOffset, desc):
        """Register field information."""
        # String items.
        self.name = name
        self.rw = rw
        self.desc = desc
        # Rest integer items.
        self.addr = addr
        self.bitTotal = bitTotal
        self.bitWidth = bitWidth
        self.bitOffset = bitOffset
        self.valueMin = valueMin
        self.valueMax = valueMax
        self.valueOffset = valueOffset
        # Generated items.
        self.bitMaskTotal = (1 << bitTotal) - 1
        self.bitMask = ( (1 << bitWidth) - 1 ) << bitOffset
        self.bitMaskNot = ( ~ self.bitMask ) & self.bitMaskTotal


class RegFields(object):
    """Register fields data-base.
    Information stored as following,
    - field name
    - read/write flag
    - reg addr
    - bit total in the whole registers, usually multiple of 8.
    - bit width
    - bit offset
    - minimum value
    - maximum value
    - offset value
    : description string
    This information is loaded from a file, usually 'regMappingFile.txt'.
    This file should be automatically dumped by the script holding the master register information.
    """
    # Class global attributes.

    def __init__(self, mappingFileName = None):
        """Initialisation."""
        # Start dictionary.
        self._fields = {}
        # Select mapping file.
        if not mappingFileName:
            self.mappingFileName = MAPPING_FILE_NAME
        # Load mapping file.
        self._loadRegisterMappingFromFile(self.mappingFileName)

    def _loadRegisterMappingFromFile(self, fileName):
        """Load register mapping from file."""
        try:
            fp = open(fileName, "r")
            lines = fp.readlines()
            fp.close()
            print("RegFields> Read register mapping from file", fileName)
        except:
            print("RegFields> Failed to open file: ", fileName)
            raise

        for line in lines:
            parts = line.split(":")
            syms = parts[0].split()
            # Remove '\n' from desc.
            desc = parts[1].strip()
            field = RegField( syms[0], syms[1], int(syms[2]), int(syms[3]), int(syms[4]), int(syms[5]), int(syms[6]), int(syms[7]), int(syms[8]), desc )
            self._fields[ syms[0] ] = field

    def getField(self, name):
        """Get reg field by name."""
        try:
            field = self._fields[name]
        except KeyError:
            raise RegFieldError( "Unknown register field name, '" + str(name) + "'." )
        except:
            raise
        return field


################################################################################
# Run if required.
def test():
    """Simple test."""
    reg = RegFields()
    for name in ('BGTRIM', 'JUNKREG'):
        print("Look for register field:", name)
        try:
            f = reg.getField(name)
            print('  Got:', f.name)
        except:
            if name in ('JUNKREG'):
                print("  Got expected error for", name)
            else:
                raise

if __name__ == "__main__":
    test()
