#!/usr/bin/env python
"""Communication using Robot USB ISS module.

NB. All serial data is bytes or bytearray.

Supports,
1) Digital input & output control.
2) Analog input.
3) I2C and/or serial.
4) SPI.

Note the USB-ISS pins have been given names IO 3, 2, 1, 0 to conform with normal digital bus vector naming convention, as opposed to the documented naming convention IO 4, 3, 2, 1.

Pins,
IO3 - I2C SDA   - SPI SDI (MISO)
IO2 - I2C SCL   - SPI SCK
IO1 - Serial Tx - SPI SDO (MOSI)
IO0 - Serial Rx - /CE

LINUX: /dev/ttyACM0 will mounted and owner by root.
To access as root do, chmod ugo+rw /dev/ttyACM0
Or more correctly, on Fedora it is owned by a group dialout.
Join this groups for correct access as a normal user.
NB. Install arduino will set this up for you.

The interface does not change the mode at start up, unless required, if it is already setup correctly.
This prevents the pins changing state during setup.

Author: Ian Watson (igw)
Date: 11 Jan 2017
"""

import sys
import os
import time
import logging

import serial
from comsutils import chrs2ints, ints2chrs, ints2hexs, value2ints, dec2bin, dec2hex, neatBinStr


################################################################################
# Logger.
LOGGER_NAME = "RobotUSBISS"
LOGGER_FILENAME = os.path.basename(__file__).split(".")[0] + time.strftime("%Y%b%d") + ".log"


# Class global attributes.
ROBOT_USBISS_HW_ID = 7
ROBOT_USBISS_FW_LATEST_VER = 0x07

# Command prefix.
CMD_USBISS = 0x5A
# Sub commands.
# CMD_USBISS + CMDS_ISS_VERSION, returns 3 bytes (module ID, firmware, operating mode).
CMDS_ISS_VERSION = 0x01
# CMD_USBISS + CMDS_ISS_MODE + ..., sets operating mode. Returns 2 status bytes.
CMDS_ISS_MODE = 0x02
# CMD_USBISS + CMDS_GET_SER_NUM, returns 8 bytes.
CMDS_GET_SER_NUM = 0x03

# Operating modes:-
# Set unused pins I/O mode, which is defined by following byte, IO_TYPE.
MODE_IO = 0x00

# Change I/O pin mode without changing operating mode.
MODE_IO_CHANGE = 0x10

MODES_IO = (MODE_IO, MODE_IO_CHANGE)

# Set I2C mode & speed. Pins IO3(SDA) & IO2(SCL).
MODE_I2C_S_20KHZ = 0x20
MODE_I2C_S_50KHZ = 0x30
MODE_I2C_S_100KHZ = 0x40
MODE_I2C_S_400KHZ = 0x50
MODE_I2C_H_100KHZ = 0x60
MODE_I2C_H_400KHZ = 0x70
MODE_I2C_H_1000KHZ = 0x80
MODES_I2C = (MODE_I2C_S_20KHZ, MODE_I2C_S_50KHZ, MODE_I2C_S_100KHZ, MODE_I2C_S_400KHZ, MODE_I2C_H_100KHZ, MODE_I2C_H_400KHZ, MODE_I2C_H_1000KHZ)

# SPI mode. All 4 pins.
MODE_SPI_0 = 0x90
MODE_SPI_1 = 0x91
MODE_SPI_2 = 0x92
MODE_SPI_3 = 0x93
MODES_SPI = (MODE_SPI_0, MODE_SPI_1, MODE_SPI_2, MODE_SPI_3)

# Serial mode. Pins IO1(Tx) & IO0(Rx). This mode can be BIT OR'ed with I2C_xyz and I assume MODES_IO.
MODE_SERIAL = 0x01

# A combined modes of I2C + SERIAL exists. This is handled as a special case and does not have a mode of itself.
#MODE_COMBI_ = 0

# Description of all modes. NB. Bit mask required for combined MODES_SPI and MODE_SERIAL.
MODE_DESCRIPTIONS = {MODE_IO : "MODE_IO", MODE_IO_CHANGE : "MODE_IO_CHANGE", MODE_I2C_S_20KHZ : "MODE_I2C_S_20KHZ", MODE_I2C_S_50KHZ : "MODE_I2C_S_50KHZ", MODE_I2C_S_100KHZ : "MODE_I2C_S_100KHZ", MODE_I2C_H_100KHZ : "MODE_I2C_H_100KHZ", MODE_I2C_H_400KHZ : "MODE_I2C_H_400KHZ", MODE_I2C_H_1000KHZ : "MODE_I2C_H_1000KHZ", MODE_SPI_0 : "MODE_SPI_0", MODE_SPI_1 : "MODE_SPI_1", MODE_SPI_2 : "MODE_SPI_2", MODE_SPI_3 : "MODE_SPI_3", MODE_SERIAL : "MODE_SERIAL"}

# IO types. 4 bit pairs for IO 4, 3, 2, 1.
IO_TYPE_O_LOW = 0
IO_TYPE_O_HIGH = 1
IO_TYPE_I = 2
IO_TYPE_I_A = 3
IO_TYPES = (IO_TYPE_O_LOW, IO_TYPE_O_HIGH, IO_TYPE_I, IO_TYPE_I_A)

# Mode (and other) returns. Return 2 bytes, CMDR_USBISS_ACK/CMDR_USBISS_NOTACK + error_code. For CMDR_USBISS_ACK, error code is always 0.
CMDR_USBISS_ACK = 0xFF
CMDR_USBISS_NOTACK = 0x00
CMDR_USBISS_I2CWRITEFAILED = 0x00

# Mode errors.
MODE_ACK_ERROR_DESCRIPTIONS = { 0x00 : "No error", 0x05 : "Unknown command", 0x06 : "Internal Error 1", 0x07 : "Internal Error 2" }

# I2C.
I2C_SGL = 0x53
I2C_AD0 = 0x54
I2C_AD1 = 0x55
I2C_AD2 = 0x56
I2C_DIRECT = 0x57

# Direct commands.
I2C_START = 0x01
I2C_RESTART = 0x02
I2C_STOP = 0x03
I2C_NACK = 0x04
# Reads 1-16 bytes, 0x20-0x2F
I2C_READ = 0x20 - 1
# Writes 1-16 bytes, 0x30-0x3F
I2C_WRITE = 0x30 - 1

# I2C errors.
I2C_ACK_ERROR_DESCRIPTIONS = { 0x01 : "Device Error. No ACK from device.", 0x02 : "Buffer Overflow. Limit frame to < 60 bytes.", 0x03 : "Buffer Underflow. More write data was expected to be sent.", 0x04 : "Unknown command. Probably your write count is wrong." }

# I/O mode controls. See IO_TYPE_x.
# Set active digital outputs. Bit must be defined as IO_TYPE_Ox. IO_SET_PINS + bit_vals, bit_vals = 0bxxxx3210. Return 1 byte, CMDR_USBISS_ACK or CMDR_USBISS_NOTACK.
IO_SET_PINS = 0x63
# Get active digital inputs. Bit must be defined as IO_TYPE_I. Return 1 byte, 0bxxxx3210.
IO_GET_PINS = 0x64
# Get active ADC input. IO_GET_ADC + pin, where pin=4, 3, 2, 1. Return 2 byte, MSB + LSB. A none ADC channel returns 0x0000, allegedly.
IO_GET_ADC = 0x65

# Serial mode controls.
CMD_SERIAL_IO = 0x62

# SPI mode controls.
CMD_SPI_IO = 0x61


class RobotUSBISSError(Exception):
    """RobotUSBISS Exception class."""
    pass


class RobotUSBISS(object):
    """Robot USB ISS coms.
    The module uses an ASCII serial stream via a serial port send over USB (windows COMx port or linux serial /dev/ttyACM0).
    Data is sent at full USB speed, not at the specified serial rates.
    Attributes,
    - mode - See MODE_ attributes.
    - baud_rate - See setMode().
    - i2cAddr - Default I2C address. Can be set at any time.
    - emulation - As it says.
    """
    
    def __init__(self, resource_name = None, mode = None, baud_rate = None, i2cAddr = 0x50, addrInc = 1, emulation = False):
        """Initialisation.
        - i2cAddr - Default 7 bit I2C address.
        - addrInc - Increment register address on each write/read. Usually 1. Mainly for emulation.
        - emulation - Prevent coms being started. Function write/read provide emulation feature.
        Unit initialises in MODE_I2C_S_100KHZ and remaining inputs in tri-state (digital inputs).
        """
        # Logger.
        self._startLogger()
        self.logger.info("Initialising Robot I2C device.")
        
        # Coms port.
        self._port_file = "robotusbiss_port.txt"
        if resource_name == None:
            resource_name = self.getPreferredPort()
        self._port = resource_name
        self.keepPreferredPort()
        # Default I2C address.
        self.i2cAddr = i2cAddr
        # For emulation, value added to regAddr after each write.
        self.regInc = addrInc
        # Emulation mode.
        self.emulation = emulation
        
        # Current mode. A mirror of the hardware defaults.
        self.mode = MODE_I2C_S_100KHZ
        # Marks pins used by specific modes. Set list element to 1 to mark when pin used by current mode. IO: 0, 1, 2, 3.
        self._pin_mode_mask = [0, 0, 1, 1]
        # Set list element with state of each IO: 0, 1, 2, 3. IO now default to tri-State (digital input).
        self._pin_io_mode_state = [IO_TYPE_I, IO_TYPE_I, IO_TYPE_I, IO_TYPE_I]
        # Actual bit mask used. IO: 0, 1, 2, 3. self._pin_io_mode_bitmask.
        self.makeIOtypeBitMask()
        # Serial. Cache buffer for serial RX.
        self.serial_rx_buffer = bytes()
        
        # Start.
        if emulation:
            self._sp = None
        else:
            self._startComs()
            # Change mode if required.
            if mode is not None:
                self.setMode(mode, baud_rate)    
                # Show the mode has changed.
                self.getHWversion()
        
        self.logger.info("  Well that was nice. All done.")
    
    def _startLogger(self):
        """Logger for robotusbiss commands."""
        self.logger = logging.getLogger(LOGGER_NAME)
        self.logger.setLevel(logging.DEBUG)
        # Create file handler which logs even debug messages.
        fh = logging.FileHandler(LOGGER_FILENAME)
        #fh.setLevel(logging.DEBUG)
        fh.setLevel(logging.INFO)
        # Create console handler with a higher log level.
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        # Create formatter and add it to the handlers.
        formatter = logging.Formatter("%(asctime)s: %(name)s: %(levelname)s: %(message)s")
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)
        # Add the handlers to the logger.
        self.logger.addHandler(fh)
        self.logger.addHandler(ch)
        
        self.logger.info("")
        self.logger.info("START logging for %s.", sys.argv[0])
    
    def getPreferredPort(self):
        """Determine the preferred port. This is stored robotusbiss_port.txt file in the run directory.
        If the file does not exist a default port is used.
        TODO: If the port in the prefered port (stored in the file) seem unsuitable, it is dropped and the default port used.
        """
        # Try to get preferred port from file.
        try:
            fd = open(self._port_file, "r")
            port = fd.readline()
            fd.close()
            port = port.strip()
        except:
            port = None
        
        # Use default port if none obtained.
        if port is None:
            port = self.getDefaultPort()
        
        return port
    
    def getDefaultPort(self):
        if sys.platform == "win32":
            return "COM5"
        return "/dev/ttyACM0"
    
    def keepPreferredPort(self):
        """Save current port to file, if possible."""
        # Attempt to save port to file.
        try:
            fd = open(self._port_file, "w")
            fd.write(self._port + os.linesep)
            fd.close()
            self.logger.info("Preferred port saved to, %s", self._port_file)
        except:
            self.logger.error("Failed to save preferred port to, %s", self._port_file)
    
    def _startComs(self):
        """Talk to robot device. Check serial number, etc."""
        self.logger.info("Opening serial comms to robot USB-ISS.")
        # Open 'COMx'. Default 9600, 8, N, 1. Set high baudrate, just in case this sets delays etc.
        self._sp = serial.Serial(self._port, 115200, timeout = 1.5)
        self.logger.info("Serial port opened: %s", self._port)
        
        # Check version.
        hwv = self.getHWversion()
        if hwv[0] != ROBOT_USBISS_HW_ID:
            raise RobotUSBISSError("HW Id not 0x07. This is not a Robot USBISS.")
        if hwv[1] <= 0x04:
            self.logger.warn("FIRMWARE VERSION POOR. Strongly suggest update to version %d.", ROBOT_USBISS_FW_LATEST_VER)
        if hwv[1] < ROBOT_USBISS_FW_LATEST_VER:
            self.logger.info("Firmware not latest. Suggest update to version %d.", ROBOT_USBISS_FW_LATEST_VER)
        
        # Note serial number.
        self.getHWsernum()
    
    def __del__(self):
        """Close coms upon delete."""
        #self.logger.info("__del__ RobotUSBISS serial.")
        self.close()
    
    def close(self):
        """Close coms port."""
        if self._sp:
            #self.logger.info("Close RobotUSBISS serial.")
            self._sp.close()
            self._sp = None
    
    def _sendCmdBytes(self, values):
        """Send data to USB-ISS.
        values - bytes or bytearray.
        """
        # Emulation.
        if self.emulation:
            self.logger.debug("USB-ISS emulation _sendCmdBytes() 0x%s", values.hex() )
            return
        
        self.logger.debug("USB-ISS _sendCmdBytes() 0x%s", values.hex() )
        self._sp.write( values )
    
    def _receiveCmdBytes(self, n):
        """Return n bytes from USB-ISS."""
        # Emulation.
        if self.emulation:
            self.logger.debug("USB-ISS emulation _receiveCmdBytes() %d", n)
            return bytes( range(n) )
        
        #self.logger.debug("Enter _receiveCmdBytes()")
        #time.sleep(0.002)
        
        #self.logger.debug("_receiveCmdBytes getting %d bytes.",  n)
#         count = 0
#         while self._sp.in_waiting < n:
#             time.sleep(0.001)
#             count += 1
#             if count > 5000:
#                 nums = self._sp.read(self._sp.in_waiting)
#                 self.logger.debug("ABORT got %d bytes", self._sp.in_waiting)
#                 break
#             self.logger.debug("WAITING FOR BUFFER TO FILL %d of %d", self._sp.in_waiting, n)
#         if count:
#             self.logger.debug("GOT BUFFER %d of %d", self._sp.in_waiting, n)
        
#         if n != self._sp.in_waiting:
#             self.logger.debug("ISSUE BUFFER HAS DIFFERENT NUMBER THAN REQUESTED %d != %d", n, self._sp.in_waiting)
        
        # Serial read is blocking till n bytes read.
        nums = self._sp.read(n)
        self.logger.debug("Exit _receiveCmdBytes() 0x%s", nums.hex() )
        return nums
    
    def getHWversion(self):
        """Get USB-ISS version and mode information (CMDS_ISS_VERSION). 3 bytes. [0]=0x07, [1]=firmware version, [2]=mode."""
        #self.logger.info("Get CMDS_ISS_VERSION, firmware,mode")
        self._sendCmdBytes( bytes( [CMD_USBISS, CMDS_ISS_VERSION] ) )
        nums = self._receiveCmdBytes(3)
        self.logger.info("HW version %d-%d. Mode %s. 0x%s", nums[0], nums[1], self.getModeDescription(nums[2]), nums.hex() )
        return nums
    
    def getModeDescription(self, mode):
        """Text description of mode."""
        if (mode & 0x01) == MODE_SERIAL and (mode & 0xF0) in MODES_I2C:
            desc = MODE_DESCRIPTIONS.get(mode & 0xF0, "Unknown USB ISS mode1") + "+" + MODE_DESCRIPTIONS.get(mode & 0x01, "Unknown USB ISS mode2")
        else:
            desc = MODE_DESCRIPTIONS.get(mode, "Unknown USB ISS mode")
        
        return desc
    
    def getHWsernum(self):
        """Get USB-ISS serial number (CMDS_GET_SER_NUM)."""
        #self.logger.info("Get CMDS_GET_SER_NUM")
        self._sendCmdBytes( bytes( [CMD_USBISS, CMDS_GET_SER_NUM] ) )
        nums = self._receiveCmdBytes(8)
        #self.logger.info("HW serial number %s, %s", ints2chrs(nums), nums.hex() )
        self.logger.info("HW serial number %s", nums)
        return nums
    
    def setMode(self, mode, baud_rate = None):
        """Set USB-ISS mode. See modes.
        - mode - One of MODE_IO, MODE_IO_CHANGE, MODES_I2C, MODE_SERIAL, MODES_SPI and combined MODES_I2C+MODE_SERIAL
            Combined I2C & Serial can be called together by, MODES_I2C | MODE_SERIAL.
        - baud_rate - The baud_rate. Either Serial, or SPI clock, SCK.
            Serial, baud_rate = any number from 300 to 1e6.
            Usually from [300, 1200, 2400, 9600, 19200, 38400, 57600, 115200, 2500000, 1000000]
            Give integer rate = (3MHz/baud_rate)-1. baud_rate = 3MHz/(rate+1)
            SPI, baud_rate = any number from 23437 to 3e6.
            Usually rate from [3MHz, 2MHz, 1.5MHz, 1.2MHz, 1MHz. 300KHz, 100KHz]
            Give integer rate = (6MHz/baud_rate)-1. baud_rate = 6MHz/(rate+1)
        
        I assume MODE_IO sets no mode, i.e. all pins are general IO.
        Then MODE_IO_CHANGE allows change of control bits.
        Use MODE_IO to clear pin usage.
        """
        self.logger.debug("Set USB-ISS mode=%s, pin_io_mode_bitmask=%s",  hex(mode), bin(self._pin_io_mode_bitmask) )
        
        # Ensure we catch SPI using all 4 pins and I2C + Serial options in correct order..
        if mode == MODE_IO:
            # Set plane IO mode. Define pins.
            self.mode = mode
            self._pin_mode_mask = [0, 0, 0, 0]
            # FIXME: Call self.makeIOtypeBitMask() ?
            self._sendCmdBytes( bytes( [CMD_USBISS, CMDS_ISS_MODE, mode, self._pin_io_mode_bitmask] ) )
        
        elif mode == MODE_IO_CHANGE:
            # Change pin IO modes.
            self._sendCmdBytes( bytes( [CMD_USBISS, CMDS_ISS_MODE, mode, self._pin_io_mode_bitmask] ) )
        
        elif mode in MODES_SPI:
            # SPI: IO3 - SPI SDI, IO2 - SPI SCK, IO1 - SPI SDO, IO0 - /CE
            self.mode = mode
            self._pin_mode_mask = [1, 1, 1, 1]
            rate = self._rateSPI(baud_rate)
            self._sendCmdBytes( bytes( [CMD_USBISS, CMDS_ISS_MODE, mode, rate] ) )
        
        elif mode in MODES_I2C:
            # I2C: IO2 - SCL, IO3 - SDA.
            self.mode = mode
            self._pin_mode_mask[2:4] = [1, 1]
            self._sendCmdBytes( bytes( [CMD_USBISS, CMDS_ISS_MODE, mode, self._pin_io_mode_bitmask] ) )
        
        elif mode == MODE_SERIAL:
            # Serial: IO0 - RX, IO1 - TX.
            self.mode = mode
            self._pin_mode_mask[0:2] = [1, 1]
            rate = self._rateSerial(baud_rate)
            self._sendCmdBytes( bytes( [CMD_USBISS, CMDS_ISS_MODE, mode, (rate & 0xFF00) >> 8, rate & 0xFF, self._pin_io_mode_bitmask] ) )
        
        elif (mode & 0xF0) in MODES_I2C and (mode & 0x0F) == MODE_SERIAL:
            # I2C and Serial.
            self.mode = mode
            self._pin_mode_mask = [1, 1, 1, 1]
            rate = self._rateSerial(baud_rate)
            self._sendCmdBytes( bytes( [CMD_USBISS, CMDS_ISS_MODE, mode, (rate & 0xFF00) >> 8, rate & 0xFF] ) )
        
        else:
            raise RobotUSBISSError("setMode unknown mode.")
        
        # Check return. Form (status, error).
        status = self._receiveCmdBytes(2)
        #self.logger.debug("setMode status=%s", status.hex() )
        
        if status[0] == CMDR_USBISS_NOTACK:
            errMes = MODE_ACK_ERROR_DESCRIPTIONS.get( status[1], "Unknown error" )
            raise RobotUSBISSError("setMode received CMDR_USBISS_NOTACK: " + errMes)
        
        if status[0] != CMDR_USBISS_ACK:
            raise RobotUSBISSError("Unknown MODE response status returned.")
        
        return True
    
    def _rateSerial(self, baud_rate):
        """Compute serial rate from baud_rate number.
        Returns integer.
        """
        if baud_rate == None:
            baud_rate = 9600
        if not ( 300 <= baud_rate <= 1000000 ):
            raise RobotUSBISSError("setMode MODE_SERIAL baud_rate out of range")
        #rate = int( ( 48.0e6 / (16 * baud_rate) ) - 0.5 )
        #baud_rate_actual = 48.0e6 / ( 16 * (rate + 1) )
        rate = int( ( 3.0e6 / baud_rate ) - 0.5 )
        baud_rate_actual = 3.0e6 / (rate + 1)
        if abs(baud_rate_actual - baud_rate) >= 1:
            self.logger.info("MODE_SERIAL baud_rate requested %d, allocated %d", baud_rate, baud_rate_actual)
        return rate
    
    def _rateSPI(self, baud_rate):
        """Compute SPI rate from baud_rate number.
        Returns integer.
        """
        if baud_rate == None:
            baud_rate = 3000000
        if not ( 23437 <= baud_rate <= 3000000 ):
            raise RobotUSBISSError("setMode MODE_SPI baud_rate out of range")
        rate = int( 6.0e6 / baud_rate - 0.5 )
        baud_rate_actual = 6.0e6 / (rate + 1)
        if abs(baud_rate_actual - baud_rate) >= 1:
            self.logger.info("MODE_SPI baud_rate requested %d, allocated %d", baud_rate, baud_rate_actual)
        return rate
    
    def makeIOtypeBitMask(self):
        """Generates an IO TYPE bit mask from self._pin_io_mode_state.
        The bit mask is stored in self._pin_io_mode_bitmask.
        Order IO3, 2, 1, 0.
        """
        self._pin_io_mode_bitmask = 0
        #23 for state, bs in zip( self._pin_io_mode_state, list(range(0, len(self._pin_io_mode_state))) ):
        for state, bs in zip( self._pin_io_mode_state, range(0, len(self._pin_io_mode_state)) ):
            self._pin_io_mode_bitmask |= state << 2*bs
    
    def setIOmode(self, pin, io_type):
        """Set pin IO type. Requested pin must be be free - not already allocated to I2C, Serial or SPI (which uses then all!).
        - pin - 0, 1, 2, 3.
        - io_type - digital out 0 or 1, digital input, analog input.
        """
        # Emulation.
        if self.emulation:
            self.logger.debug("setIOmode emulation")
            return
        self.logger.debug("setIOmode")
        if io_type not in IO_TYPES:
            raise RobotUSBISSError("Given IO_TYPE not valid.")
        if self._pin_mode_mask[pin]:
            raise RobotUSBISSError("Requested pin not a general IO. Already in use by I2C/SERIAL/SPI mode.")
        self._pin_io_mode_state[pin] = io_type
        self.makeIOtypeBitMask()
        self.setMode(MODE_IO_CHANGE)
    
    def setIOpins(self, pin_state):
        """Set state of all active digital output pins. Digital or analog input pins, I2C pins, or serial pins are unaffected.
        - pin_state - Pin as bit mask. Only bottom 4 bits used - mask 0bxxx4321.
        """
        # Emulation.
        if self.emulation:
            self.logger.debug("setIOpins emulation")
            return CMDR_USBISS_ACK
        self._sendCmdBytes( bytes( [IO_SET_PINS, pin_state] ) )
        # Check return status.
        status = self._receiveCmdBytes(1)[0]
        if status == CMDR_USBISS_NOTACK:
            raise RobotUSBISSError("setIOpins failed.")
        if status != CMDR_USBISS_ACK:
            raise RobotUSBISSError("Unknown response status returned.")
        return status
    
    def setIOpin(self, pin, pin_state):
        """Set ONLY digital output pin values.
        - pin - pin number 0, 1, 2, 3.
        - pin_state - 0 or 1.
        """
        # Get current pin values and mask in the pin bit.
        pinvals = ( self.getIOpins() & ( 0xFF & ~(1 << pin) ) ) | ( (pin_state & 1) << pin )
        return self.setIOpins(pinvals)
    
    def getIOpins(self):
        """Get values of ALL digital outputs and inputs. Mask 0bxxx4321."""
        # Emulation.
        if self.emulation:
            self.logger.debug("getIOpins emulation")
            return 0
        self._sendCmdBytes( bytes( [IO_GET_PINS] ) )
        pinvals = self._receiveCmdBytes(1)
        return pinvals[0]
    
    def getIOpin(self, pin):
        """Get values of specfic digital IO output or input.
        - pin - pin number 0, 1, 2, 3.
        """
        pinval = ( self.getIOpins() >> pin ) & 1
        return pinval
    
    def getAD(self, ch):
        """Get ADC input. Value 0-1023 inclusive. Inactive analog inputs return 0x0000.
        - ch - 0, 1, 2, 3
        """
        if ch < 0 or ch > 3:
            raise RobotUSBISSError("ADC channels 0 to 3 only.")
        if self.emulation:
            self.logger.debug("getAD emulation")
            return 0
        self._sendCmdBytes( bytes( [IO_GET_ADC, ch+1] ) )
        adcval = self._receiveCmdBytes(2)
        adcval = ( adcval[0] << 8 ) | adcval[1]
        return adcval
    
    
    ############################################################################
    # Following are I2C write and read to various address length formats.
    
    def sendI2CbytesA1(self, regAddr, regValues, i2cAddr):
        """Send number of bytes to register (addr 1 byte wide), of I2C addr device.
        Sequential multi-byte writes requires auto regAddr increment/decrement for correct operation.
        - regAddr - int8 register address.
        - regValues - bytes, bytearray or uint8 list.
        - i2cAddr - 7 bit I2C address.
        Return status. 0 OK.
        """
        regValues = bytes(regValues)
        self.logger.debug("Enter sendI2CbytesA1() A%d R%d -> 0x%s", i2cAddr, regAddr, regValues.hex() )
        # Convert to 8 bit format with bit 0 = '0' for write.
        i2cAddr = ( ( i2cAddr << 1 ) & 0xFE ) + 0
        if len(regValues) > 60:
            err_mess = "To many values to write - max 60."
            self.logger.error(err_mess)
            raise RobotUSBISSError(err_mess)
        cmds = bytes( [I2C_AD1, i2cAddr, regAddr, len(regValues)] ) + regValues
        self.logger.debug("sendI2CbytesA1 Send CMDS.")
        self._sendCmdBytes(cmds)
        # Check return. Form: status.
        self.logger.debug("sendI2CbytesA1 Get status.")
        status = self._receiveCmdBytes(1)
        #status = [0x00]
        if status[0] == CMDR_USBISS_I2CWRITEFAILED:
            self.logger.error("sendI2CbytesA1 write FAILED.")
            #raise RobotUSBISSError("sendI2CbytesA1 write failed.")
        #else:
        #    self.logger.debug("sendI2CbytesA1 write OK.")
        self.logger.debug("Exit sendI2CbytesA1() %d", status[0] )
        return status[0]
    
    def receiveI2CbytesA1(self, regAddr, n, i2cAddr):
        """Receiver number of bytes from register (addr 1 byte wide), of I2C addr device.
        Sequential multi-byte writes requires auto regAddr increment/decrement for correct operation.
        This does i2c write + regAddr + restart + read's.
        - regAddr - int8 register address.
        - n - number bytes to get.
        - i2cAddr - 7 bit I2C address.
        Return bytes.
        """
        self.logger.debug("Enter receiveI2CbytesA1() A%d R%d <- %d", i2cAddr, regAddr, n )
        # Convert to 8 bit format with bit 0 = '1' for read.
        i2cAddr = ( ( i2cAddr << 1 ) & 0xFE ) + 1
        if n > 60:
            err_mess = "To many values to read - max 60."
            self.logger.error(err_mess)
            raise RobotUSBISSError(err_mess)
        self.logger.debug("receiveI2CbytesA1 Send CMDS.")
        self._sendCmdBytes( bytes( [I2C_AD1, i2cAddr, regAddr, n] ) )
        # Get read.
        self.logger.debug("receiveI2CbytesA1 Get reads.")
        regValues = self._receiveCmdBytes(n)
        if len(regValues) != n:
            self.logger.error("receiveI2CbytesA1 read FAILED.")
            #raise RobotUSBISSError("receiveI2CbytesA1 read failed.")
        self.logger.debug("Exit receiveI2CbytesA1() 0x%s", regValues.hex() )
        return regValues
    
    def receiveI2CbytesA1b(self, regAddr, n, i2cAddr):
        """PLAYING.
        Receiver number of bytes from register (addr 1 byte wide) of I2C addr device.
        Assumes regAddr will be incremented/decrements after each write.
        Read cycle using custom format: S+W+Waddr+P + S+Rdata+P, as apposed to normal 2 cycle read, S+W+Raddr+P + S+Wread+Rdata+P.
        - regAddr - int8 register address.
        - n - number bytes to get.
        - i2cAddr - 7 bit I2C address.
        Return bytes.
        """
        self.logger.debug("Enter receiveI2CbytesA1b A%d R%d <- %d", i2cAddr, regAddr, n )
        # Convert to 8 bit format with bit 0 = '1' for read.
        #i2cAddr = ( ( i2cAddr << 1 ) & 0xFE ) | 1
        i2cAddr = ( ( i2cAddr << 1 ) & 0xFE )
        if n > 16:
            err_mess = "To many values to read - max 16."
            self.logger.error(err_mess)
            raise RobotUSBISSError(err_mess)
        #cmds = [I2C_DIRECT, I2C_START, I2C_WRITE+2, i2cAddr | 0, regAddr, I2C_STOP, I2C_START, I2C_WRITE+1, i2cAddr | 1, I2C_NACK, I2C_READ+1, I2C_STOP]
        cmds = bytes( [I2C_DIRECT, I2C_START, I2C_WRITE+2, i2cAddr | 0, regAddr, I2C_STOP, I2C_START, I2C_WRITE+1, i2cAddr | 1, I2C_READ+1, I2C_STOP] )
        #if n > 1:
        #    cmds = cmds + [I2C_READ + n - 1]
        #cmds = cmds + [I2C_START, I2C_NACK, I2C_READ + 1, I2C_STOP]
        self.logger.debug("receiveI2CbytesA1b Send CMDS.")
        self._sendCmdBytes(cmds)
        # Check return. Form: [status, error, numbers ...).
        self.logger.debug("receiveI2CbytesA1b Get status+reads")
        regValues = self._receiveCmdBytes(2+n)
        if regValues[0] == CMDR_USBISS_NOTACK:
            errMes = I2C_ACK_ERROR_DESCRIPTIONS.get(regValues[1], "Unknown error")
            raise RobotUSBISSError("receiveI2CbytesA1b received CMDR_USBISS_NOTACK: " + errMes)
        if regValues[0] != CMDR_USBISS_ACK:
            raise RobotUSBISSError("Unknown CMDR_USBISS_ACK status returned.")
        if regValues[1] != n:
            raise RobotUSBISSError("Read of " + str(n) + " bytes only returned " + str(regValues[1]))
        regValues = regValues[2:]
        self.logger.debug("receiveI2CbytesA1b() 0x%s", regValues.hex() )
        return regValues
    
    def receiveI2CbytesA1_2(self, regAddr, n, i2cAddr):
        """Receiver number of bytes from register (addr 1 byte wide) of I2C addr device.
        Sequential multi-byte writes requires auto regAddr increment/decrement for correct operation.
        Read cycle using custom format: S+W+Waddr+Rdata+P, as opposed to normal 2 cycle read, S+W+Raddr+P + S+Wread+Rdata+P.
        - regAddr - int8 register address.
        - n - number bytes to get.
        - i2cAddr - 7 bit I2C address.
        Return bytes.
        """
        self.logger.debug("Enter receiveI2CbytesA1_2 A%d R%d <- %d", i2cAddr, regAddr, n )
        # Convert to 8 bit format with bit 0 = '1' for read.
        i2cAddr = ( ( i2cAddr << 1 ) & 0xFE ) | 1
        if n > 16:
            err_mess = "To many values to read - max 16."
            self.logger.error(err_mess)
            raise RobotUSBISSError(err_mess)
        cmds = [I2C_DIRECT, I2C_START, I2C_WRITE+2, i2cAddr, regAddr]
        if n > 1:
            cmds.extend( [I2C_READ + n - 1] )
        cmds.extend( [I2C_NACK, I2C_READ + 1, I2C_STOP] )
        self.logger.debug("receiveI2CbytesA1_2 Send CMDS.")
        self._sendCmdBytes( bytes(cmds) )
        # Check return. Form: [status, error, numbers ...).
        self.logger.debug("receiveI2CbytesA1_2 Get status+reads")
        regValues = self._receiveCmdBytes(2+n)
        if regValues[0] == CMDR_USBISS_NOTACK:
            errMes = I2C_ACK_ERROR_DESCRIPTIONS.get(regValues[1], "Unknown error")
            raise RobotUSBISSError("receiveI2CbytesA1_2 received CMDR_USBISS_NOTACK: " + errMes)
        if regValues[0] != CMDR_USBISS_ACK:
            raise RobotUSBISSError("Unknown CMDR_USBISS_ACK status returned.")
        if regValues[1] != n:
            raise RobotUSBISSError("Read of " + str(n) + " bytes only returned " + str(regValues[1]))
        regValues = regValues[2:]
        self.logger.debug("receiveI2CbytesA1_2() 0x%s", regValues.hex() )
        return regValues
    
    
    ############################################################################
    # Following are serial write and read.
    
    def sendReceiveSerialBytes(self, values = bytes(), n = True):
        """Send number of bytes to serial TX. It automatically return any in the RX buffer.
        An error will occur if more than 32 are send at once. You may send no values, which will return any waiting RX values.
        Any call update the serial_txcount.
        - values - bytes, bytearray or uint8 list. 0 to 32 off, 0 being bytes(0).
        - n - Number of RX items. True for all.
        Return bytes.
        """
        values = bytes(values)
        self.logger.debug("Enter sendReceiveSerialBytes() -> 0x%s", values.hex() )
        if len(values) > 32:
            err_mess = "Serial sends a maximum of 32 values."
            self.logger.error(err_mess)
            raise RobotUSBISSError(err_mess)
        self.logger.debug("sendReceiveSerialBytes Send CMDS.")
        self._sendCmdBytes( bytes([CMD_SERIAL_IO]) + values )
        # Check return. Form: TX CMDR_USBISS_ACK/NACK, TX sent count, RX ready count.
        self.logger.debug("sendReceiveSerialBytes Get status.")
        status, self.serial_txcount, rxcount = self._receiveCmdBytes(3)
        if status == CMDR_USBISS_NOTACK:
            err_mess = "sendReceiveSerialBytes write failed."
            self.logger.error(err_mess)
            raise RobotUSBISSError(err_mess)
        if status != CMDR_USBISS_ACK:
            raise RobotUSBISSError("Unknown CMDR_USBISS_ACK status returned.")
        if rxcount > 0:
            serial_rx = self._receiveCmdBytes(rxcount)
        else:
            serial_rx = bytes()
        self.logger.debug("Exit sendReceiveSerialBytes() TX %d RX %d", self.serial_txcount, rxcount )
        self.serial_rx_buffer += serial_rx
        if n == True:
            serial_rx = self.serial_rx_buffer
            self.serial_rx_buffer = bytes()
        else:
            serial_rx = self.serial_rx_buffer[0:n]
            self.serial_rx_buffer = self.serial_rx_buffer[n:]
        return serial_rx
    
    def sendSerialBytes(self, values):
        """Send number of bytes to serial TX. See sendReceiveSerialBytes.
        Any waiting RX are pulled into a buffer.
        """
        serial_rx = self.sendReceiveSerialBytes(values)
        # Keep any received character in buffer.
        self.serial_rx_buffer += serial_rx
    
    def receiveSerialBytes(self, n = True):
        """Receive n bytes from serial RX. See sendReceiveSerialBytes.
        Items in buffer are prepended.
        n - Number of RX items. True for all.
        Return bytes.
        """
        serial_rx = self.sendReceiveSerialBytes(n = n)
        return serial_rx
    
    
    ############################################################################
    # Following is SPI write and read. NB It reads at the same time as the write.
    
    def sendReceiveSPIbytes(self, values):
        """Send number of bytes to SPI, during which it will receives the same number of bytes.
        An error will occur if more than 62 are send at once.
        - values - bytes, bytearray or uint8 list. 1 to 62 off.
        Return bytes.
        """
        if not values:
            return bytes()
        values = bytes(values)
        self.logger.debug("Enter sendReceiveSPIbytes() -> 0x%s", values.hex() )
        if len(values) < 1 or len(values) > 62:
            err_mess = "SPI sends between 1 to 62 values."
            self.logger.error(err_mess)
            raise RobotUSBISSError(err_mess)
        self.logger.debug("sendReceiveSPIbytes Send.")
        self._sendCmdBytes( bytes([CMD_SPI_IO]) + values )
        # Check return. Form: TX ACK/NACK, RX data (same length as TX).
        self.logger.debug("sendReceiveSPIbytes Receive.")
        spi_rx = self._receiveCmdBytes( 1 + len(values) )
        if spi_rx[0] == CMDR_USBISS_NOTACK:
            err_mess = "sendReceiveSPIbytes write failed."
            self.logger.error(err_mess)
            raise RobotUSBISSError(err_mess)
        if spi_rx[0] != CMDR_USBISS_ACK:
            raise RobotUSBISSError("Unknown CMDR_USBISS_ACK status returned.")
        return spi_rx[1:]
