#!/usr/bin/env python
"""Instrument VISA communications interface.
This is a common interface to all instrument.
It handles: GPIB, USB, RS232(ASRL), ethernet, and direct pyserial COM's.

It is basically a wrapper around PyVISA. It was done as a wrapper so that GPIB devices may also be controlled by a 'prologix' GPIB controller. In effect to select between two controllers. 
It also distinguishes ASRL names, as serial COM ports under VISA, from normal COM ports, which will be handled by pyserial. Simple instruments, like the Robot ones, using simple binary bytes sent over serial.

PyVISA-1.4 was being used. SVN 126. This had no ResourceManager, just a call to open the instrument. Also message termination stripping occurred on reads.

Change to use latest, PyVISA-1.7, at time.
On fish,
rm = visa.ResourceManager()
rm.list_resources()
rm
<ResourceManager(<VisaLibrary('C:\\Windows\\system32\\visa32.dll')>)>

VISA drivers.
Context click on 'Aglient IO Control' icon. This tells you where visa libraries are located.

On fish. Agilent visa files in,
C:\\Windows\\system32\\visa32.dll
C:\\Windows\\system32\\visa32.Agilent Technologies.dll
C:\\Windows\\system32\\visaext.dll

On rb11. Agilent visa files in,
C:\Program Files\Agilent\IO Libraries Suite\intfcfg\visacf32.dll
C:\Program Files\IVI Foundation\VISA\WinNT\lib\bc\visa...
C:\Program Files\IVI Foundation\VISA\WinNT\lib\msc\visa...various.

Backends can be seen by command line,
python -m visa info

The command line provides a interactive shell
python -m visa shell
Examples,
help
list
open 0
query *IDN?
close 0

Termination.
GPIB:- Most writes are '\r\n' (CR+LF + EOI). Most reads are '\r\n' (CR+LF + EOI).
Serial:- Most are '\r' (CR). Most reads are '\r' (CR).

Author: Ian Watson (igw)
Date: 21 Jul 2015

FIXME: Change all the COM serial stuff to use the term_chars like pyvisa. For now we do most serial COM though the pyvisa ASRL, which works well.
"""

import time

import visa
# Debug logging.
#visa.log_to_screen()
import visaprologix
import serial


################################################################################
# Globals
def reverse_name_map(name_map):
    """Create a reverse name map dictionary."""
    return dict( [ (v, k) for k, v in name_map.items() ] )

def bidirectional_name_map(name_map):
    """Appends reverse of dictionary mapping."""
    for key, value in list(name_map.items()):
        name_map[value] = key
    return name_map


# Name of args known to VisaComsBase.
VISA_COMS_ARG_NAMES = ("emulation", "GPIBcontroller", "port", "prompt", "echo", "skip_idn", "what_am_i", "read_term_chars_strip", "use_wait_opc", "write_delay", "read_delay", "freq_unit", "level_unit")

# SCPI boolean map.
SCPI_MAP_BOOL_TO_NUM_BOOL = { "OFF" : "0", "ON" : "1" }
SCPI_MAP_BOOL_TO_TEXT_BOOL = { "0" : "OFF", "1" : "ON" }

# Write/read fixed delays.
DELAY_AFTER_WRITE = 0
DELAY_BEFORE_READ = 0


################################################################################
class VisaComsBaseError(Exception):
    """VisaComsBase Exception class."""
    pass


################################################################################
# Global resource manager.
resourceManagerVisaGlobal = None
    
class VisaComsBase(object):
    """Base class for instruments using VISA/GPIB interface."""
    
    def __init__(self, resource_name = None, **keyw):
        """
        Arguments for visacomsbase:-
        - resource_name - Format either,
          a) 'GPIBx::y' , x="", 0, 1, etc., y=int 0-31.
             or 'USB0', 'USBInstrument1', usually virtual serial port over USB, but a VISA TSM interface.
             or 'COMx', direct access to serial COM ports, usually virtual serial port over USB. x=integer. Uses pyserial.
          b) IP/hostname address 'A.B.C.D' or 'abc.local.net' name format; or "COMn" for prologix unit.        
        - GPIBcontroller - prologix | visa. Either prologix or visa controller (agilent, NI, etc).
        - port - prologix COM port name.
        - emulation - Emulation mode. No actual coms channel. True/False.
        - echo - echo commands, for debugging. Default False.
        - prompt - echo prompt, for debugging. Set '' for no prompt. Default 'Unknown'.
        - skip_idn - Skip *IDN? printout. True/False. Default does ieee_idn().
        - what_am_i - Set by inheriting class. String. Description in emulation mode.
        - read_term_chars_strip - Termination characters to strip from end of each message read.
        - use_wait_opc - Use OPC? after each command. True/False. Default False.
        - write_delay - Delay after each write. Default 0.
        - read_delay - Delay before each read. Default 0.
        - freq_unit - Frequency units used. Default ''.
        - level_unit - Level units used. Default ''.
        
        Arguments for pyvisa and others:-
        - The rest of arguments are passed to visa (pyvisa) and serial (pyserial).
        """
        global resourceManagerVisaGlobal
        
        # 'resource_name' is names to be similar to pyvisa interface.
        if resource_name is not None:
            self.resource_name = resource_name
        
        # Create local copy of class args.
        if not hasattr(self, "args_visa"):
            self.args_visa = {}
        
        if not hasattr(self, "args_visacoms"):
            self.args_visacoms = {}
        
        # Strip arguments into args_visacoms and args_visa parts.
        for key, value in keyw.items():
            if key in VISA_COMS_ARG_NAMES:
                self.args_visacoms[key] = value
            else:
                self.args_visa[key] = value
        
        # Set default visacoms parameters.
        self.emulation = self.args_visacoms.get("emulation", False)
        # HACK Comment here to globally use prologix or agilent.
        self.GPIBcontroller = self.args_visacoms.get("GPIBcontroller", "visa")
        #self.GPIBcontroller = self.args_visacoms.get("GPIBcontroller", "prologix")
        # port just passed on within args_visacoms.
        # Console echo & prompt.
        self.echo = self.args_visacoms.get("echo", False)
        self.prompt = self.args_visacoms.get("prompt", "Unknown")
        # Skip *IDN? command & identity.
        self.skip_idn = self.args_visacoms.get("skip_idn", False)
        self.what_am_i = self.args_visacoms.get("what_am_i", "Unknown VISA device.")
        # Read end string termination chars. These will be stripped automatically.
        self.read_term_chars_strip = self.args_visacoms.get("read_term_chars_strip", None)
        # A delay after each write & read.
        self.write_delay = self.args_visacoms.get("write_delay", 0)
        self.read_delay = self.args_visacoms.get("read_delay", 0)
        # Use OPC? command to wait for completion of last command.
        self.use_wait_opc = self.args_visacoms.get("use_wait_opc", False)
        
        # Frequency units for SpecAna & SigGen.
        self.freq_unit = self.args_visacoms.get("freq_unit", "")
        # Level units for SpecAna & SigGen.
        self.level_unit = self.args_visacoms.get("level_unit", "")
        
        # Open instrument.
        if self.emulation:
            self.instrument = None
        
        else:
            # Open ResourceManager. Create just one.
            if resourceManagerVisaGlobal is None:
                resourceManagerVisaGlobal = visa.ResourceManager()
                # List available instruments.
                #print "D-VisaComsBase>Resources_list> ", resourceManagerVisaGlobal.list_resources()
                res_info = resourceManagerVisaGlobal.list_resources_info()
                print("D-VisaComsBase>Resources_list> ", list(res_info.keys()))
                for v in res_info.values():
                    print("D-VisaComsBase>Resources> Name = %s, Alias = %s" % ( v[3], v[4] ))
            
            if self.resource_name[:4] == "GPIB":
                # Use prologix, otherwise default to pyvisa for Agilent, NI, etc, GPIB controller.
                if self.GPIBcontroller == "prologix":
                    self.instrument = visaprologix.instrument(self.resource_name, **self.args_visacoms)
                else:
                    self.instrument = resourceManagerVisaGlobal.open_resource(self.resource_name, **self.args_visa)
            
            elif self.resource_name[:3] == "COM":
                # Open 'COMx' with pyserial. Default 2400, 8, N, 1.
                # NB. Use ASRL__ notation to access COM port by visa interface.
                self.instrument = serial.Serial(self.resource_name, **self.args_visa)
            
            else:
                # Not GPIB or Serial COM, so it will be USB, RS232 COM's via visa ASRL__ name (rather than COM), or ethernet, which pyvisa will handle or return an error.
                self.instrument = resourceManagerVisaGlobal.open_resource(self.resource_name, **self.args_visa)
        
        # Print feedback of device ID request.
        if self.skip_idn:
            print("D-VisaComsBase> *IDN? DISABLED:", self.what_am_i)
        elif not self.emulation:
            self.what_am_i = self.ieee_idn_get()
        print("D-VisaComsBase> *IDN?", self.what_am_i)
    
    
    ############################################################################
    # HACK code. Uncomment this code to check things are being closed correctly.
    #def __del__(self):
    #    """Close visacoms upon delete."""
    #    print("D-VisaComsBase> __del__ - calling close()", self.resource_name)
    #    self.close()

    #def close(self):
    #    """Close visacoms port."""
    #    print("D-VisaComsBase> Close - which does nothing.", self.resource_name)
    #    #self.instrument.close()
    
    
    ############################################################################
    # SCPI write & read.
    def write(self, message, delay = True):
        """Write SCPI message."""
        if self.echo or self.emulation:
            print("D-VisaComsBase> %s> write( '%s')" % ( self.prompt, message.rstrip() ))
        if self.emulation:
            return True

        # For serial via pyserial append '\n'.
        if self.resource_name[0:3] == "COM":
            message += "\n"
            self.instrument.write( message.encode() )
        else:
            self.instrument.write(message)
        
        if delay and self.write_delay + DELAY_AFTER_WRITE > 0:
            time.sleep(self.write_delay + DELAY_AFTER_WRITE)
    
    def read(self, delay = True, debug = "0"):
        """Read SCPI message."""
        if self.emulation:
            print("D-VisaComsBase> %s> read() = '%s'" % ( self.prompt, str(debug) ))
            return debug
        
        if delay and self.read_delay + DELAY_BEFORE_READ > 0:
            time.sleep(self.read_delay + DELAY_BEFORE_READ)
        
        # For serial via pyserial. Strip trailing '\n'.
        if self.resource_name[0:3] == "COM":
            mess = self._readSerial()
            mess = mess.rstrip()
        else:
            mess = self.instrument.read()
        
        # Strip any end termination characters. These can be any way round, like \n\r or \r\n.
        if self.read_term_chars_strip:
            mess = mess.rstrip(self.read_term_chars_strip)
        if mess and mess[-1] in ['\n', '\r']:
            print("D-VisaComsBase> WARNING read still has termination chars at end of string.")
            time.sleep(5)
        
        if self.echo:
            print("D-VisaComsBase> %s> read() = '%s'" % ( self.prompt, str(mess) ))
        
        return mess
    
    def _readSerial(self):
        """Read from the serial buffer until end or time out.
        The end will be or should be \r\n or just \n.
        So just keep reading until \n found.
        Return the full message - with termination.
        TODO: Try to change to read in bytes until \n found.
        """
        # Delay should not be required here. 
        #time.sleep(0.1)
        mess = bytearray()
        count = 0
        while True:
            buf = self.instrument.in_waiting
            if buf > 0:
                mess.extend( self.instrument.read(buf) )
            #print "D-VisaComsBase> read(%d) mess = '%s'" % ( buf, mess.replace('\r', '\\r').replace('\n', '\\n') )
            if len(mess) > 1 and mess[-2:] == b"\n*":
                mess = mess[:-1]
            if len(mess) > 0 and mess[-1] == b"\n":
                break
            #print "D-VisaComsBase> Waiting for buffer to fill - count=%d, buf=%d" % (count, buf)
            if count >= 25:
                print("D-VisaComsBase> Something wrong, timeout reached!")
                raise VisaComsBaseError("VisaComBase Serial COM timed out.")
            count += 1
            time.sleep(0.1)
        return mess.decode()
    
    
    ############################################################################
    # GPIB wait.
    def wait(self, opc = True):
        """Wait until SCPI commands have completed.
        opc - True/False do an OPC wait.
        """
        if not opc:
            return "1"
        #print "D-VisaComsBase> *OPC? Checking if SCPI commands complete."
        t0 = time.time()
        count = 0
        # 1000 * 0.1s = 100 seconds. At least with overheads.
        while count < 1000:
            # Ask if all previous commands have been completed.
            self.write("*OPC?", delay = False)
            mess = self.read(delay = False, debug = "1")
            #print "D-VisaComsBase> *OPC? :", mess
            # Might return '*OPC 1' or '1', so use -1 index to check value.
            if str( mess.split()[-1] ) != "0":
                break
            if count < 1:
                print("D-VisaComsBase> *OPC? Waiting until SCPI commands complete.")
            count += 1
            time.sleep(0.1)
        if count > 0:
            print("D-VisaComsBase> *OPC? waited %d loops, for %.3f s." % ( count, time.time() - t0 ))
        return mess
    
    
    ############################################################################
    # Multi purpose write/read functions.
    def get_query(self, scpimess, debug = "0"):
        """Convert SCPI command into query and get response.
        e.g., :STATE to :STATE? and returns its response."""
        self.write( scpimess + "?" )
        # Don't see need to .wait() for a query. The buffer it trying to respond with the reply.
        mess = self.read(debug = debug)
        return mess
    
    def arg1_with_query(self, scpimess = "", asserts = None, mode = None, query = False, debug = "0"):
        """Quick SCPI command with 1 argument. Argument is space (' ') separated from the SCPI command.
        opc - True/False local OPC wait override.
        """
        if query:
            return self.get_query(scpimess, debug)
        if asserts is not None:
            assert mode in asserts, "Not known"
        if mode is not None:
            scpimess = scpimess + " " + str(mode)
        self.write( scpimess )
        # If instrument supports it, wait until command completes.
        self.wait(self.use_wait_opc)
        return None
    
    def cmd1_with_query(self, scpimess = "", asserts = None, mode = None, query = False, debug = "0"):
        """Quick command with 1 command argument. Argument is colon (':') append to the main SCPI command.
        e.g., Given CONF with VOLT:DC gives, CONF:VOLT:DC
        If a query, only the main command is queried.
        e.g., CONF?
        opc - True/False local OPC wait override.
        """
        if query:
            return self.get_query(scpimess, debug)
        if asserts is not None:
            assert mode in asserts, "Not known"
        if mode is not None:
            scpimess = scpimess + ":" + str(mode)
        self.write( scpimess )
        # If instrument supports it, wait until command completes.
        self.wait(self.use_wait_opc)
        return None
    
    
    ############################################################################
    # General IEEE commands.
    def ieee_idn_get(self):
        """IEEE *IDN?
        Identification query.
        """
        return self.get_query("*IDN", debug = "Debug IDN.")
    
    def ieee_opc_get(self):
        """IEEE *OPC?
        Returns 1 when previous command has completed.
        """
        return self.get_query("*OPC", debug = "1")
    
    def ieee_rst(self, seconds = 3):
        """IEEE *RST
        Reset the instrument. Waits till ready - simple delay.
        """
        self.write("*RST")
        time.sleep(seconds)
    
    def ieee_trg(self):
        """IEEE *TRG
        Trigger.
        """
        self.write("*TRG")
    
    def ieee_esr_get(self):
        """IEEE *ESR?
        Query and clear event status register. To determine errors.
        """
        mess = self.get_query("*ESR")
        return int(mess)
    
    
    ############################################################################
    # Utility.
    def sleep_em(self, seconds):
        """time.sleep for emulation"""
        if not self.emulation:
            time.sleep(seconds)
    
    ############################################################################
    # Argument mapping.
    def map2num_bool(self, tag):
        """Map SCPI boolean keywords towards numerical boolean.
        So 0 or OFF to 0, 1 or ON to 1.
        All items returned as str.
        """
        if tag is None:
            return tag
        # Ensure tag is string.
        return SCPI_MAP_BOOL_TO_NUM_BOOL.get( str(tag), str(tag) )
    
    def map2text_bool(self, tag):
        """Map SCPI boolean keywords towards text boolean.
        So 0 or OFF to OFF, 1 or ON to ON.
        All items returned as str.
        """
        if tag is None:
            return tag
        # Ensure tag is string.
        return SCPI_MAP_BOOL_TO_TEXT_BOOL.get( str(tag), str(tag) )
