#!/usr/bin/env python
"""Prologix GPIB-USB module.
This Prologix unit provides a GPIB controller & device interface over USB, via an FTDI serial port.
Not intelligent, but works.

Methods and attributes, as "pyvisa" (for those implemented).
Note.
To match the pyvisa interface, all write messages require no \n at end, and all reads will have any terminating \n stripped off.
So the prologix write commands add \n, and the read strips \n.

LINUX: /dev/ttyUSB0 will mount as root. For user access, as root do,
chmod ugo+rw /dev/ttyUSB0

All commands must be terminated by CR=13='\r' or LF=10='\n'.
Commands:-
++addr 0-30 [96-126]  -- specify GPIB address 
++addr                -- query GPIB address 
++auto 0|1            -- enable (1) or disable (0) read-after-write 
++auto                -- query read-after-write setting 
++clr                 -- issue device clear 
++eoi 0|1             -- enable (1) or disable (0) EOI with last byte 
++eoi                 -- query eoi setting 
++eos 0|1|2|3         -- EOS terminator - 0:CR+LF, 1:CR, 2:LF, 3:None 
++eos                 -- query eos setting 
++eot_enable 0|1      -- enable (1) or disable (0) appending eot_char on EOI 
++eot_enable          -- query eot_enable setting 
++eot_char <char>     -- specify eot character in decimal 
++eot_char            -- query eot_char character 
++ifc                 -- issue interface clear 
++loc                 -- set device to local 
++lon                 -- enable (1) or disable (0) listen only mode 
++mode 0|1            -- set mode - 0:DEVICE, 1:CONTROLLER 
++mode                -- query current mode 
++read [eoi|<char>]   -- read until EOI, <char>, or timeout 
++read_tmo_ms 1-3000  -- set read timeout in millisec
++read_tmo_ms         -- query timeout 
++rst                 -- reset controller 
++savecfg 0|1         -- enable (1) or disable (0) saving configuration to EPROM 
++savecfg             -- query savecfg setting 
++spoll               -- serial poll currently addressed device 
++spoll 0-30 [96-126] -- serial poll device at specified address 
++srq                 -- query SRQ status 
++status 0-255        -- specify serial poll status byte 
++status              -- query serial poll status byte 
++trg                 -- issue device trigger 
++ver                 -- query controller version 
++help                -- display this help

Author: Ian Watson (igw)
Date: 9 Feb 2012
"""

import sys
import time

import serial
# For functions.
import robotusbiss


################################################################################
# The following hold the interface to each controller. Each controller is referenced by GPIBx::, where x is an integer.
# As this module looks like 'pyvisa', then controller interface is hidden and accessed only by the 'GPIBx::' format. This is a simple cludge to achieve this.
# Only one interface, 'GPIB::'. This could be a list of interfaces, 'GPIB0::', 'GPIB1::', etc.
controller_interface = None


################################################################################
class PrologixUSBControllerError(Exception):
    """PrologixUSBController Exception class."""
    pass


class PrologixInstrumentError(Exception):
    """PrologixInstrument Exception class."""
    pass


def parse_name_gpib(name):
    """Parse GPIB name of form 'GPIBx::n'. x defaults to 0.
    The short name 'GPIB' will be taken as 'GPIB0'.
    Ex. 'GPIB::12' --> ['GPIB0', '12']."""
    syms = name.split("::")
    if syms[0][0:4] != "GPIB":
        raise PrologixUSBControllerError("GPIB resource name must begin 'GPIB'.")
    if syms[0] == "GPIB":
        syms[0] = "GPIB0"
    return syms


def initialise_controller(resource_name = "GPIB0", port = None):
    """Assigned controller to interface resource name & number, i.e., GPIB0.
    The short name 'GPIB' will be taken as 'GPIB0'.
    Return controller_interface so its easy to close."""
    global controller_interface
    
    gpib_resource_addr = parse_name_gpib(resource_name)
    controller_interface = PrologixUSBController(gpib_resource_addr[0], port)
    print("D-PxUSBVisa> Started controller for", gpib_resource_addr[0])
    return controller_interface


def close_controller(resource_name = None):
    """Close controller. If none specified, closes ALL controller.
    As there is only one. Hey ho."""
    global controller_interface
    controller_interface.close()


STR_EOS = { "0" : "CR+LF", "1" : "CR", "2" : "LF", "3" : "None" }

class PrologixUSBController(object):
    """Prologix USB-GPIB VISA controller.
    The module uses an ASCII coms stream over a virtual serial port send over USB.
    The data rate at USB speed, not serial limited.
    - resource_name - 'GPIBx::y'
    - port - windows COMx port or linux serial /dev/ttyUSB0. Generally change the COM port to match default.
    """

    def __init__(self, resource_name, port = None):
        """Initialisation."""
        # GPIB resource name.
        self.resource_name = resource_name
        # Current GPIB address.
        self.addr = None
        # Coms port.
        if port == None:
            if sys.platform == "win32":
                port = "COM13"
            else:
                port = "/dev/ttyUSB0"
        self._port = port
        # Start coms.
        self._startComs()

    def _startComs(self):
        """Initialise prologix hardware. Check serial number, etc."""
        print("D-PxUSBVisa> Initialising Prologix USB-GPIB controller.")
        
        print("D-PxUSBVisa> Opening serial comms to Prologix over USB.")
        # Open 'COMx'. Default 9600, 8, N, 1. Use higher baudrate just in case it is limiting the throughput.
        self._sp = serial.Serial(self._port, baudrate=115200, timeout = 2.5)
        print("D-PxUSBVisa> Serial port opened:", self._port)
        
        # This delay gives chance for the prologix unit to process its input buffer. Otherwise the next write will overwrite it. Experiment shows 0.3 work with HP 34404 better.
        self.prologix_write_delay = 0.3
        
        # Reset.
        #self.sendCmd("++rst\n")
        #time.sleep(6)
        # Stop EPROM from wearing out!
        # EPROM saves: mode, addr, auto, eoi, eos, eot_enable, eot_char and read_tmo_ms.
        self.sendCmd("++savecfg\n")
        mess = self.recvCmd()
        print("D-PxUSBVisa> ++savecfg =", mess)
        if mess == "1":
            self.sendCmd("++savecfg 0\n")
        # Mode is a CONTROLLER.
        self.sendCmd("++mode 1\n")
        # Auto read-after-write. 0=off, 1=on.
        self.sendCmd("++auto 0\n")
        # Enable/disable EOI (End Of Inquire?) signal at end of send GPIB message. Some instruments require this.
        self.sendCmd("++eoi 1\n")
        # Write messaged End Of String appended characters, with, 0: CR+LF; 1:CR (13='\r'); 2: LF(10='\n'); 3:None.
        self.sendCmd("++eos 2\n")
        # Received End Of Transmission char added to whenever EOI detected when reading string from GPIB port. 0=off, 1=on.
        self.sendCmd("++eot_enable 0\n")
        self.sendCmd("++eot_enable 1\n")#XXX
        # Receive End Of Transmission char added (after bus EOI). 42=*, 10='\n'.
        self.sendCmd("++eot_char 42\n")
        # Timeout read and spoll commands. Default 500ms.
        self.sendCmd("++read_tmo_ms 3000\n")
        
        # Unit version.
        self.sendCmd("++ver\n")
        mess = self.recvCmd()
        print("D-PxUSBVisa> ++ver got '%s'" % ( mess ))
        self.sendCmd("++eoi\n")
        mess = self.recvCmd()
        print("D-PxUSBVisa> ++eoi = %s" % ( mess ))
        self.sendCmd("++eos\n")
        mess = self.recvCmd()
        print("D-PxUSBVisa> ++eos = %s '%s'" % ( mess, STR_EOS[mess] ))
        self.sendCmd("++eot_char\n")
        mess = self.recvCmd()
        print("D-PxUSBVisa> ++eot_char = %s '%s'" % ( mess, chr( int(mess) ) ))
        self.sendCmd("++read_tmo_ms\n")
        mess = self.recvCmd()
        print("D-PxUSBVisa> ++read_tmo_ms = %s" % ( mess ))
        self.sendCmd("++addr\n")
        mess = self.recvCmd()
        print("D-PxUSBVisa> ++addr = %s" % ( mess ))
        
        print("D-PxUSBVisa>   Well that was nice. All done.")

    def __del__(self):
        """Close coms upon delete."""
        #print "D-PxUSBVisa> __del__ controller - calling close()", self.resource_name
        self.close()

    def close(self):
        """Close coms port."""
        #print "D-PxUSBVisa> Close prologix controller", self.resource_name
        self._sp.close()
    
    def sendCmd(self, mess):
        """Send a command. Must start with ++. Terminating \n must be included."""
        assert mess[0:2] == "++", "Prologix command must start with '++'."
        #print "D-PxUSBVisa> sendCmd(%s) > %s" % (self.resource_name, mess.rstrip())
        self._sp.write(mess)
        # This flush probably not required, but cannot harm.
        self._sp.flush()
        # Small delay to let proligix unit process command just send. IF another command is send before completion, it overwrites it input buffer. See prologix document, section 6.2. Experiment shows 0.01 fails, but 0.02 works. Safety margin.
        time.sleep(self.prologix_write_delay)
    
    def recvCmd(self):
        """Receive the response from a command.
        Return message with trailing \r and/or \n removed."""
        mess = self.readUntilEnd()
        #print "D-PxUSBVisa> recvCmd()", mess.rstrip(), robotusbiss.chrs2ints(mess)
        mess = mess.rstrip()
        return mess
    
    def setGPIBaddr(self, addr):
        """Set the GPIB address to talk to.
        - addr - int or str."""
        # Only change GPIB address if it has changed.
        if addr == self.addr:
            return
        self.addr = addr
        # Ensure only one terminating \n.
        mess = "++addr " + str(addr).rstrip() + "\n"
        self.sendCmd(mess)
    
    def sendGPIB(self, addr, mess):
        """Send message to GPIB device with given address.
        - addr - int or str.
        - mess - string
        """
        #print "D-PxUSBVisa> sendGPIB(%s::%s) %s" % (self.resource_name, str(addr), mess.rstrip())
        # Set GPIB address.
        self.setGPIBaddr(addr)
        # Send message to GPIB device.
        self._sp.write(mess)
        # This flush probably not required, but cannot harm.
        self._sp.flush()
        # See sendCmd() above..
        time.sleep(self.prologix_write_delay)
    
    def recvGPIB(self):
        """Read message fromGPIB device. Trailing \n is stripped off.
        The last written GPIB device, as in address, is used."""
        # It seems a sendGPIB followed by recvGPIB, that the _sp.write, followed by the _sp.read(++read) is causing the issue. So a delay, either here help. Being here it does not slow down sendGPIB throughput.
        #xxxxxxtime.sleep(0.5)
        # Read until EOI detected or timeout (read_tmo_ms). Could read until 10=LF='\n'.
        self.sendCmd("++read eoi\n")
        #self.sendCmd("++read 10\n")
        #self.sendCmd("++read\n")
        # A delay need was found unnecessary.
        #time.sleep(0.5)
        mess = self.readUntilEnd()
        #print "D-PxUSBVisa> recvGPIB(%s) %s" % (self.resource_name, mess.rstrip())
        # Message returned with trailing \n. Strip to match pyvisa.
        mess = mess.rstrip()
        return mess
    
    def readUntilEnd(self):
        """Read from the bufer until end or time out.
        The end will be or should be \r\n or just \n.
        So just keep reading until \n found.
        Return the full message - with termination."""
        # Delay should not be required here. 
        #time.sleep(0.1)
        mess = ""
        count = 0
        while True:
            buf = self._sp.in_waiting
            if buf > 0:
                mess += self._sp.read(buf)
            #print "D-PxUSBVisa> read(%d) mess = '%s'" % ( buf, mess.replace("\r", "\\r").replace("\n", "\\n") )
            #print "  which is,", robotusbiss.chrs2ints(mess)
            if len(mess) > 1 and mess[-2:] == "\n*":
                mess = mess[:-1]
            if len(mess) > 0 and mess[-1] == "\n":
                break
            #print "D-PxUSBVisa> Waiting for buffer to fill - count=%d, buf=%d" % (count, buf)
            if count >= 25:
                print("D-PxUSBVisa> Something wrong, timeout reached!")
                raise PrologixUSBControllerError("Prologix readUntilEnd timed out.")
                #break
            count += 1
            time.sleep(0.2)
        #print "D-PxUSBVisa> readUntilEnd()", mess.rstrip(), robotusbiss.chrs2ints(mess)
        return mess


def instrument(resource_name, **keyw):
    """Factory function for instrument instances.
    - resource_name - the VISA resource name of the device.  It may be an alias.
    - keyw - keyword argument for the class constructor of the device instance to be generated.  See the class Instrument for further information.
    Return value:
    The generated instrument instance.
    """
    instrument = PrologixInstrument(resource_name, **keyw)
    return instrument


class PrologixInstrument(object):
    """Prologix VISA instrument. This looks like pyvisa interface.
    All write & read IO are calls the attached controller."""
    
    def __init__(self, resource_name, **keyw):
        """Constructor method.
        - resource_name -- the instrument's resource name or an alias, may be taken from the list from get_instruments_list().
        Keyword arguments:
        port -- the prologix serial coms port "COMx"
        from pyvisa
        timeout -- the VISA timeout for each low-level operation in milliseconds.
        term_chars -- the termination characters for this device, see description of class property "term_chars".
        chunk_size -- size of data packets in bytes that are read from the device.
        lock -- whether you want to have exclusive access to the device. Default: VI_NO_LOCK
        delay -- waiting time in seconds after each write command. Default: 0
        send_end -- whether to assert end line after each write command. Default: True
        values_format -- floating point data value format.  Default: ascii (0)
        """
        global controller_interface
        # Get resource name and address.
        self.resource_name, self.gpib_addr = parse_name_gpib(resource_name)
        # Get the controller for this resource name. Currently we only handle one controller.
        # TODO: Find the controller for this GPIB resource, or create one for it.
        if not controller_interface:
            port = keyw.get("port") # .get() defaults to None.
            initialise_controller(self.resource_name, port = port)
        # So we have created controller, or it already existed.
        self.controller = controller_interface
        # Ensure the one and only controller is assigned to the same resource, i.e., GPIB0.
        if self.resource_name != self.controller.resource_name:
            raise PrologixInstrumentError("No controller initialised for resource " + self.resource_name)
        print("D-PxUSBVisa> Controller to instrument %s::%s ready" % (self.resource_name, self.gpib_addr))

    def write(self, message):
        """As pyvisa instrument.write(). No terminating character required.
        Note, pyvisa has no \n at end. The prologix requires this.
        """
        # Make controller point at correct GPIB address.
        self.controller.sendGPIB(self.gpib_addr, message.rstrip() + "\n")

    def read(self):
        """As pyvisa instrument.read(). No terminating characters returned."""
        mess = self.controller.recvGPIB()
        return mess

    # Other pyvisa methods not implemented.
